package org.artifactory.config;

import lombok.AccessLevel;
import lombok.NoArgsConstructor;
import org.apache.commons.lang3.SerializationUtils;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.descriptor.config.MutableCentralConfigDescriptor;
import org.artifactory.descriptor.property.PropertySet;
import org.artifactory.descriptor.replication.LocalReplicationDescriptor;
import org.artifactory.descriptor.replication.RemoteReplicationDescriptor;
import org.artifactory.descriptor.repo.FederatedRepoDescriptor;
import org.artifactory.descriptor.repo.KeyPairDescriptor;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.ProxyRepoDescriptor;
import org.artifactory.descriptor.repo.RemoteRepoDescriptor;
import org.artifactory.descriptor.repo.RepoLayout;
import org.artifactory.descriptor.repo.VirtualRepoDescriptor;
import org.artifactory.descriptor.repo.releasebundle.ReleaseBundlesRepoDescriptor;
import org.artifactory.descriptor.security.SecurityDescriptor;
import org.artifactory.descriptor.security.ldap.LdapSetting;
import org.artifactory.descriptor.security.ldap.group.LdapGroupSetting;
import org.artifactory.descriptor.security.oauth.OAuthSettings;
import org.artifactory.descriptor.security.sso.CrowdSettings;
import org.artifactory.descriptor.security.sso.SamlSettings;
import org.artifactory.descriptor.sumologic.SumoLogicConfigDescriptor;

import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import static org.artifactory.repo.system.SystemRepos.SystemRepoConfig.isImportAllowed;

/**
 * This class performs several sanitizations of the confgi descriptor
 * Should be used only for SAAS migration process
 *
 * @author Igor Usenko
 */
@NoArgsConstructor(access = AccessLevel.PRIVATE)
public class CentralConfigDescriptorSanitizer {

    public static MutableCentralConfigDescriptor merge(CentralConfigDescriptor source,
            MutableCentralConfigDescriptor target) {
        MutableCentralConfigDescriptor result = SerializationUtils.clone(target);
        mergeConfigDescriptors(source, result);
        return result;
    }

    private static void mergeConfigDescriptors(CentralConfigDescriptor source, MutableCentralConfigDescriptor result) {
        mergeLocalRepos(source, result);
        mergeRemoteRepos(source, result);
        mergeFederatedRepos(source, result);
        mergeVirtualRepos(source, result);
        mergeReleaseBundles(source, result);
        mergeSecurityDescriptors(source, result);
        mergePropertySets(source, result);
        mergeRepoLayouts(source, result);
        mergeRemoteReplications(source, result);
        mergeLocalReplications(source, result);
        mergeSumoLogicConfigs(source, result);
        mergeKeyPairs(source, result);

        result.setDateFormat(source.getDateFormat());
        result.setGlobalReplicationConfig(SerializationUtils.clone(source.getReplicationsConfig()));
        result.setHelpLinksEnabled(source.isHelpLinksEnabled());
        result.setLogo(source.getLogo());
        result.setSystemMessageConfig(SerializationUtils.clone(source.getSystemMessageConfig()));
        result.setTrashcanConfig(SerializationUtils.clone(source.getTrashcanConfig()));
        result.setReleaseBundlesConfig(SerializationUtils.clone(source.getReleaseBundlesConfig()));
        result.setSignedUrlConfig(SerializationUtils.clone(source.getSignedUrlConfig()));
        result.setEulaConfig(SerializationUtils.clone(source.getEulaConfig()));
        result.setSubscriptionConfig(SerializationUtils.clone(source.getSubscriptionConfig()));
        result.setAuthentication(SerializationUtils.clone(source.getAuthentication()));
    }

    private static void mergeKeyPairs(CentralConfigDescriptor source, MutableCentralConfigDescriptor result) {
        List<KeyPairDescriptor> sourceKeyPairs = source.getKeyPairs();
        if (sourceKeyPairs != null) {
            result.setKeyPairs(sourceKeyPairs.stream().map(SerializationUtils::clone).collect(
                    Collectors.toList()));
        }
    }

    private static void mergeSumoLogicConfigs(CentralConfigDescriptor source, MutableCentralConfigDescriptor result) {
        SumoLogicConfigDescriptor sourceSumoLogicConfig = source.getSumoLogicConfig();
        if (sourceSumoLogicConfig != null) {
            SumoLogicConfigDescriptor sumoLogic = SerializationUtils.clone(sourceSumoLogicConfig);
            sumoLogic.setProxy(null);
            result.setSumoLogicConfig(sumoLogic);
        }
    }

    private static void mergeLocalReplications(CentralConfigDescriptor source, MutableCentralConfigDescriptor result) {
        List<LocalReplicationDescriptor> sourceLocalReplications = source.getLocalReplications();
        if (sourceLocalReplications != null) {
            result.setLocalReplications(sourceLocalReplications.stream().map(localReplicationDescriptor -> {
                LocalReplicationDescriptor descriptor = SerializationUtils.clone(localReplicationDescriptor);
                descriptor.setProxy(null);
                return descriptor;
            }).collect(Collectors.toList()));
        }
    }

    private static void mergeRemoteReplications(CentralConfigDescriptor source, MutableCentralConfigDescriptor result) {
        List<RemoteReplicationDescriptor> sourceRemoteReplications = source.getRemoteReplications();
        if (sourceRemoteReplications != null) {
            result.setRemoteReplications(sourceRemoteReplications.stream().map(SerializationUtils::clone).collect(
                    Collectors.toList()));
        }
    }

    private static void mergeRepoLayouts(CentralConfigDescriptor source, MutableCentralConfigDescriptor result) {
        List<RepoLayout> sourceRepoLayouts = source.getRepoLayouts();
        if (sourceRepoLayouts != null) {
            result.setRepoLayouts(sourceRepoLayouts.stream().map(SerializationUtils::clone).collect(
                    Collectors.toList()));
        }
    }

    private static void mergePropertySets(CentralConfigDescriptor source, MutableCentralConfigDescriptor result) {
        List<PropertySet> sourcePropertySets = source.getPropertySets();
        if (sourcePropertySets != null) {
            result.setPropertySets(sourcePropertySets.stream().map(SerializationUtils::clone).collect(
                    Collectors.toList()));
        }
    }

    private static void mergeSecurityDescriptors(CentralConfigDescriptor source,
            MutableCentralConfigDescriptor result) {
        SecurityDescriptor sourceSecurity = source.getSecurity();
        if (sourceSecurity != null) {
            SecurityDescriptor resultSecurity = result.getSecurity();

            // RTDEV-26275 - we are skipping httpsso in purpose as it not supported in saas
            mergeLdapSettings(sourceSecurity, resultSecurity);
            mergeLdapGroupSettings(sourceSecurity, resultSecurity);
            mergeCrowdSettings(sourceSecurity, resultSecurity);
            mergeSamlSettings(sourceSecurity, resultSecurity);
            mergeOAuthSettings(sourceSecurity, resultSecurity);

            resultSecurity.setAnonAccessEnabled(sourceSecurity.isAnonAccessEnabled());
            resultSecurity.setHideUnauthorizedResources(sourceSecurity.isHideUnauthorizedResources());
            resultSecurity.setSigningKeysSettings(SerializationUtils.clone(sourceSecurity.getSigningKeysSettings()));
            resultSecurity.setBuildGlobalBasicReadAllowed(sourceSecurity.isBuildGlobalBasicReadAllowed());
            resultSecurity.setBuildGlobalBasicReadForAnonymous(sourceSecurity.isBuildGlobalBasicReadForAnonymous());
            resultSecurity.setBasicAuthEnabled(sourceSecurity.isBasicAuthEnabled());

            result.setSecurity(resultSecurity);
        }
    }

    private static void mergeOAuthSettings(SecurityDescriptor sourceSecurity, SecurityDescriptor resultSecurity) {
        OAuthSettings sourceSecurityOauthSettings = sourceSecurity.getOauthSettings();
        if (sourceSecurityOauthSettings != null) {
            OAuthSettings oAuthSettings = SerializationUtils.clone(sourceSecurityOauthSettings);
            oAuthSettings.setEnableIntegration(false);
            resultSecurity.setOauthSettings(oAuthSettings);
        }
    }

    private static void mergeSamlSettings(SecurityDescriptor sourceSecurity, SecurityDescriptor resultSecurity) {
        SamlSettings sourceSecuritySamlSettings = sourceSecurity.getSamlSettings();
        if (sourceSecuritySamlSettings != null) {
            SamlSettings samlSettings = SerializationUtils.clone(sourceSecuritySamlSettings);
            samlSettings.setEnableIntegration(false);
            resultSecurity.setSamlSettings(samlSettings);
        }
    }

    private static void mergeCrowdSettings(SecurityDescriptor sourceSecurity, SecurityDescriptor resultSecurity) {
        CrowdSettings sourceSecurityCrowdSettings = sourceSecurity.getCrowdSettings();
        if (sourceSecurityCrowdSettings != null) {
            CrowdSettings crowdSettings = SerializationUtils.clone(sourceSecurityCrowdSettings);
            crowdSettings.setEnableIntegration(false);
            resultSecurity.setCrowdSettings(crowdSettings);
        }
    }

    private static void mergeLdapGroupSettings(SecurityDescriptor sourceSecurity, SecurityDescriptor resultSecurity) {
        List<LdapGroupSetting> sourceSecurityLdapGroupSettings = sourceSecurity.getLdapGroupSettings();
        if (sourceSecurityLdapGroupSettings != null) {
            resultSecurity.setLdapGroupSettings(
                    sourceSecurityLdapGroupSettings.stream().map(SerializationUtils::clone).collect(
                            Collectors.toList()));
        }
    }

    private static void mergeLdapSettings(SecurityDescriptor sourceSecurity, SecurityDescriptor resultSecurity) {
        List<LdapSetting> sourceSecurityLdapSettings = sourceSecurity.getLdapSettings();
        if (sourceSecurityLdapSettings != null) {
            List<LdapSetting> ldapSettings = sourceSecurityLdapSettings.stream().map(ldapSetting -> {
                LdapSetting l = SerializationUtils.clone(ldapSetting);
                l.setEnabled(false);
                return l;
            }).collect(Collectors.toList());
            resultSecurity.setLdapSettings(ldapSettings);
        }
    }

    private static void mergeReleaseBundles(CentralConfigDescriptor source, MutableCentralConfigDescriptor result) {
        Map<String, ReleaseBundlesRepoDescriptor> sourceReleaseBundlesRepositoriesMap =
                source.getReleaseBundlesRepositoriesMap();
        if (sourceReleaseBundlesRepositoriesMap != null) {
            Map<String, ReleaseBundlesRepoDescriptor> releaseBundlesRepositories = sourceReleaseBundlesRepositoriesMap
                    .entrySet().stream()
                    .filter(e -> isImportAllowed(e.getKey()))
                    .collect(Collectors.toMap(Map.Entry::getKey, e -> SerializationUtils.clone(e.getValue())));
            result.setReleaseBundlesRepositoriesMap(releaseBundlesRepositories);
        }
    }

    private static void mergeVirtualRepos(CentralConfigDescriptor source, MutableCentralConfigDescriptor result) {
        Map<String, VirtualRepoDescriptor> sourceVirtualRepositoriesMap = source.getVirtualRepositoriesMap();
        if (sourceVirtualRepositoriesMap != null) {
            Map<String, VirtualRepoDescriptor> virtualRepositories = sourceVirtualRepositoriesMap.entrySet().stream()
                    .filter(e -> isImportAllowed(e.getKey()))
                    .collect(Collectors.toMap(Map.Entry::getKey, e -> SerializationUtils.clone(e.getValue())));
            result.setVirtualRepositoriesMap(virtualRepositories);
        }
    }

    private static void mergeFederatedRepos(CentralConfigDescriptor source, MutableCentralConfigDescriptor result) {
        Map<String, FederatedRepoDescriptor> sourceFederatedRepositoriesMap = source.getFederatedRepositoriesMap();
        if (sourceFederatedRepositoriesMap != null) {
            Map<String, FederatedRepoDescriptor> federatedRepositories =
                    sourceFederatedRepositoriesMap.entrySet().stream()
                            .filter(e -> isImportAllowed(e.getKey()))
                            .collect(Collectors.toMap(Map.Entry::getKey, e -> {
                                FederatedRepoDescriptor clone = SerializationUtils.clone(e.getValue());
                                clone.setProxy(null);
                                return clone;
                            }));
            result.setFederatedRepositoriesMap(federatedRepositories);
        }
    }

    private static void mergeRemoteRepos(CentralConfigDescriptor source, MutableCentralConfigDescriptor result) {
        Map<String, RemoteRepoDescriptor> sourceRemoteRepositoriesMap = source.getRemoteRepositoriesMap();
        if (sourceRemoteRepositoriesMap != null) {
            Map<String, RemoteRepoDescriptor> remoteRepositories = sourceRemoteRepositoriesMap.entrySet().stream()
                    .filter(e -> isImportAllowed(e.getKey()))
                    .collect(Collectors.toMap(Map.Entry::getKey, e -> {
                        RemoteRepoDescriptor clone = SerializationUtils.clone(e.getValue());
                        if (clone instanceof ProxyRepoDescriptor) {
                            ((ProxyRepoDescriptor) clone).setProxy(null);
                        }
                        return clone;
                    }));
            result.setRemoteRepositoriesMap(remoteRepositories);
        }
    }

    private static void mergeLocalRepos(CentralConfigDescriptor source, MutableCentralConfigDescriptor result) {
        Map<String, LocalRepoDescriptor> sourceLocalRepositoriesMap = source.getLocalRepositoriesMap();
        if (sourceLocalRepositoriesMap != null) {
            Map<String, LocalRepoDescriptor> localRepositories = sourceLocalRepositoriesMap.entrySet().stream()
                    .filter(e -> isImportAllowed(e.getKey()))
                    .collect(Collectors.toMap(Map.Entry::getKey, e -> SerializationUtils.clone(e.getValue())));
            result.setLocalRepositoriesMap(localRepositories);
        }
    }

}
