package org.artifactory.ph;

import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.reflect.MethodUtils;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.core.api.ph.exception.PackageRepoDescriptorException;
import org.artifactory.core.api.ph.service.PackageArtifactoryConfigService;
import org.artifactory.core.api.ph.vcs.configuration.PhVcsGitConfiguration;
import org.artifactory.descriptor.repo.RemoteRepoDescriptor;
import org.artifactory.descriptor.repo.RepoBaseDescriptor;
import org.artifactory.descriptor.repo.VcsConfiguration;
import org.artifactory.descriptor.repo.VcsGitConfiguration;
import org.artifactory.repo.Repo;
import org.artifactory.repo.service.InternalRepositoryService;
import org.artifactory.util.HttpUtils;
import org.jfrog.client.util.PathUtils;

import javax.servlet.http.HttpServletRequest;

import static org.apache.http.HttpStatus.SC_BAD_REQUEST;
import static org.apache.http.HttpStatus.SC_INTERNAL_SERVER_ERROR;
import static org.apache.http.HttpStatus.SC_NOT_FOUND;
import static org.artifactory.ph.PackageRepoServiceImpl.DOES_NOT_EXIST;

/**
 * @author Ori Yafe
 */
@Slf4j
public class PackageArtifactoryConfigServiceImpl implements PackageArtifactoryConfigService {

    public static final String REMOTE_REPOSITORY = "Remote repository '";
    private final InternalRepositoryService repositoryService;
    private final CentralConfigService centralConfigService;

    public PackageArtifactoryConfigServiceImpl(InternalRepositoryService repositoryService,
            CentralConfigService centralConfigService) {
        this.repositoryService = repositoryService;
        this.centralConfigService = centralConfigService;
    }

    @Override
    public String getStringFromRepoConfig(String repoKey, String key, String def)
            throws PackageRepoDescriptorException {
        return getValue(repoKey, key, def, String.class);
    }

    @Override
    public int getIntFromRepoConfig(String repoKey, String key, int def) throws PackageRepoDescriptorException {
        return getValue(repoKey, key, def, Integer.class);
    }

    @Override
    public boolean getBooleanFromRepoConfig(String repoKey, String key, boolean def)
            throws PackageRepoDescriptorException {
        return getValue(repoKey, key, def, Boolean.class);
    }

    @Override
    public long getLongFromRepoConfig(String repoKey, String key, long def) throws PackageRepoDescriptorException {
        return getValue(repoKey, key, def, Long.class);
    }

    @Override
    public String getArtifactoryBaseUrl(HttpServletRequest httpRequest) {
        return PathUtils.trimTrailingSlashes(HttpUtils.getArtifactoryBaseUrl(httpRequest));
    }

    @Override
    public PhVcsGitConfiguration getVcsConfigOfRemoteRepo(String remoteRepoKey) throws PackageRepoDescriptorException {
        RemoteRepoDescriptor remoteRepoDescriptor = repositoryService.remoteRepoDescriptorByKey(remoteRepoKey);
        if (remoteRepoDescriptor == null) {
            String msg = REMOTE_REPOSITORY + remoteRepoKey + "' was not found";
            log.warn(msg);
            throw new PackageRepoDescriptorException(SC_NOT_FOUND, msg, msg);
        }
        VcsConfiguration vcs = remoteRepoDescriptor.getVcs();
        if (vcs == null) {
            String msg = REMOTE_REPOSITORY + remoteRepoKey + "' does not have VCS configuration";
            log.warn(msg);
            throw new PackageRepoDescriptorException(SC_BAD_REQUEST, msg, msg);
        }
        VcsGitConfiguration vcsGitConfiguration = vcs.getGit();
        if (vcsGitConfiguration == null) {
            String msg = REMOTE_REPOSITORY + remoteRepoKey + "' does not have VCS Git configuration";
            log.warn(msg);
            throw new PackageRepoDescriptorException(SC_BAD_REQUEST, msg, msg);
        }
        return PhVcsConfigurationMapper.from(vcsGitConfiguration);
    }


    private <T> T getValue(String repoKey, String key, T def, Class<T> clazz) throws PackageRepoDescriptorException {
        T result = null;
        Repo repository = repositoryService.repositoryByKey(repoKey);
        if (repository == null) {
            throw new PackageRepoDescriptorException(SC_NOT_FOUND, PackageRepoServiceImpl.NOT_FOUND,
                    "Repo with name:" + repoKey + DOES_NOT_EXIST);
        }
        RepoBaseDescriptor repoDescriptor = getRepoDescriptor(repository);
        try {
            Object fieldResults = MethodUtils.invokeMethod(repoDescriptor, key);
            result = clazz.cast(fieldResults);
        } catch (Exception e) {
            log.debug("Failed to get key: {} from repository: {}", key, repoKey, e);
            throw new PackageRepoDescriptorException(SC_INTERNAL_SERVER_ERROR, "Internal error", e.getMessage(), e);
        }
        return result != null ? result : def;
    }

    private RepoBaseDescriptor getRepoDescriptor(Repo repo) {
        RepoBaseDescriptor repoBaseDescriptor;
        String repoKey = repo.getKey();
        if (repo.isReal()) {
            if (repo.isLocal()) {
                repoBaseDescriptor =
                        centralConfigService.getDescriptor().getLocalAndFederatedRepositoriesMap().get(repoKey);
            } else {
                repoBaseDescriptor = centralConfigService.getDescriptor().getRemoteRepositoriesMap().get(repoKey);
            }
        } else {
            repoBaseDescriptor = centralConfigService.getDescriptor().getVirtualRepositoriesMap().get(repoKey);
        }
        return repoBaseDescriptor;
    }
}
