package org.artifactory.ph;

import com.google.common.annotations.VisibleForTesting;
import com.google.common.collect.Multimap;
import org.artifactory.api.properties.PropertiesService;
import org.artifactory.common.repo.PackageType;
import org.artifactory.common.repo.RepoInfo;
import org.artifactory.common.repo.RepoType;
import org.artifactory.core.api.ph.exception.PackageException;
import org.artifactory.core.api.ph.service.PackageRepoService;
import org.artifactory.core.api.ph.virtual.resolution.PackageRepo;
import org.artifactory.descriptor.repo.RepoDescriptor;
import org.artifactory.md.Properties;
import org.artifactory.repo.RemoteRepo;
import org.artifactory.repo.RealRepo;
import org.artifactory.repo.Repo;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.RepoPathFactory;
import org.artifactory.repo.service.InternalRepositoryService;
import org.artifactory.repo.virtual.VirtualRepo;
import org.artifactory.repo.virtual.VirtualRepositoriesResolver;
import org.artifactory.util.DoesNotExistException;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.List;
import java.util.stream.Collectors;

import static org.apache.http.HttpStatus.SC_INTERNAL_SERVER_ERROR;
import static org.apache.http.HttpStatus.SC_NOT_FOUND;
import static org.artifactory.ph.PackageRepoTypeMapper.packageType;

/**
 * The service contains the minimal repository services needed for the packages
 */
public class PackageRepoServiceImpl implements PackageRepoService {
    private static final Logger log = LoggerFactory.getLogger(PackageRepoServiceImpl.class);
    public static final String NOT_FOUND = "Not found";
    public static final String DOES_NOT_EXIST = " doesn't exist";
    private final InternalRepositoryService repositoryService;
    private final PropertiesService propertiesService;

    public PackageRepoServiceImpl(InternalRepositoryService repositoryService, PropertiesService propertiesService) {
        this.repositoryService = repositoryService;
        this.propertiesService = propertiesService;
    }

    @Override
    public boolean isVirtual(String repoKey) {
        return repositoryService.virtualRepositoryByKey(repoKey) != null;
    }

    @Override
    public boolean isRemote(String repoKey) {
        return repositoryService.remoteRepositoryByKey(repoKey) != null;
    }

    @Override
    public boolean isLocal(String repoKey) {
        return repositoryService.localOrFederatedRepositoryByKey(repoKey) != null;
    }

    @Override
    public void move(String sourceRepoKey, String targetRepoKey, String sourcePath, String targetPath)
            throws PackageException {
        RepoPath source = RepoPathFactory.create(sourceRepoKey, sourcePath);
        RepoPath target = RepoPathFactory.create(targetRepoKey, targetPath);
        try {
            repositoryService.move(source, target, false, true, true);
        } catch (Exception e) {
            log.error("Failed to execute move", e);
            throw new PackageException(SC_INTERNAL_SERVER_ERROR, "Internal error", e.getMessage(), e);
        }
    }

    @Override
    @NotNull
    public RepoInfo getRepoInfo(String repoKey) throws PackageException {
        Repo<?> repo = repositoryService.repositoryByKey(repoKey);
        if (repo != null) {
            if (repo.isLocal()) {
                return new RepoInfo(RepoType.LOCAL, packageType(repo));
            } else if (repo.isReal()) {
                return new RepoInfo(RepoType.REMOTE, packageType(repo));
            }
            return new RepoInfo(RepoType.VIRTUAL, packageType(repo));
        }
        String msg = String.format("Couldn't find repo info for repo:%s", repoKey);
        throw new PackageException(SC_NOT_FOUND, NOT_FOUND, msg);
    }

    @Override
    @Nullable
    public PackageType getPackageType(String repoKey) {
        Repo<?> repo = getRepo(repoKey);
        if (repo != null) {
            return packageType(repo);
        }
        return null;
    }

    @Nullable
    private Repo<?> getRepo(String repoKey) {
        Repo<?> repo = repositoryService.localOrFederatedRepositoryByKey(repoKey);
        if (repo == null) {
            repo = repositoryService.remoteRepositoryByKey(repoKey);
            if (repo == null) {
                repo = repositoryService.virtualRepositoryByKey(repoKey);
            }
        }
        return repo;
    }

    @Override
    public void setAttributes(String repoKey, String path, Multimap<String, String> properties) {
        RepoPath repoPath = RepoPathFactory.create(repoKey, path);
        Properties finalExistingProperties = propertiesService.getProperties(repoPath);
        properties.asMap().forEach((key, collection) -> {
            finalExistingProperties.removeAll(key);
            if (collection != null) {
                finalExistingProperties.putAll(key, collection);
            }
        });
        propertiesService.setProperties(repoPath, finalExistingProperties, true);
    }

    @Override
    public boolean existLocally(String repoKey, String path) {
        boolean exists;
        VirtualRepo virtualRepo = repositoryService.virtualRepositoryByKey(repoKey);
        if (virtualRepo != null) {
            exists = virtualRepo.itemExists(path);
        } else {
            // in case of local, federated, and remote cache
            exists = repositoryService.exists(RepoPathFactory.create(repoKey, path));
        }
        return exists;
    }

    @Override
    public void undeployArtifact(String repoKey, String path) {
        repositoryService.undeploy(RepoPathFactory.create(repoKey, path));
    }

    @Override
    public List<? extends PackageRepo> getSubRepositories(String repoKey,
            @Nullable String path, boolean filterCacheRepositories) throws PackageException {
        VirtualRepo virtualRepo = repositoryService.virtualRepositoryByKey(repoKey);
        if (virtualRepo != null) {
            VirtualRepositoriesResolver resolver = virtualRepo.getVirtualRepositoriesResolver();
            RepoPath repoPath = RepoPathFactory.create(repoKey, path);
            List<RealRepo<?>> res = resolver.assembleSearchRepositoriesList(repoPath, false);
            if (filterCacheRepositories) {
                res = res.stream().filter(realRepo -> !realRepo.isCache()).collect(Collectors.toList());
            }
            return res;
        }
        throw new PackageException(SC_NOT_FOUND, NOT_FOUND,
                "virtual repo with name:" + repoKey + DOES_NOT_EXIST);
    }

    @Override
    public String getDefaultDeploymentRepoKey(String repoKey) throws PackageException {
        try {
            if (repositoryService.isVirtualRepoExist(repoKey)) {
                return repositoryService.getDefaultDeploymentRepoKey(repoKey);
            }
            return repoKey;
        } catch (DoesNotExistException e) {
            throw new PackageException(SC_NOT_FOUND, NOT_FOUND, e.getMessage(), e);
        }
    }

    @Override
    public void undeployVirtualMetadataCache(String localRepoKey, String virtualRepoMetadataRoot) {
        log.info("Triggering virtual metadata cache eviction for local repository '{}' for path '{}'",
                localRepoKey, virtualRepoMetadataRoot);
        try {
            getVirtualReposContainingRepo(localRepoKey)
                    .forEach(repo -> evictVirtualCache(repo.getKey(), virtualRepoMetadataRoot));
        } catch (PackageException e) {
            log.error("Failed to retrieve virtual repositories list containing repo '{}' with an error: {}",
                    localRepoKey, e.getPublicReason());
        }
    }

    @Override
    public String getRemoteRepoBaseUrl(String remoteRepoKey) throws PackageException {
        RemoteRepo<?> remoteRepo = repositoryService.remoteRepositoryByKey(remoteRepoKey);
        if (remoteRepo == null) {
            String message = "Remote repo " + remoteRepoKey + " does not exist";
            throw new PackageException(SC_NOT_FOUND, message, message);
        }
        return remoteRepo.getUrl();
    }

    /**
     * Check if index exists in virtual repo. Index could never been created or already evicted.
     * if exists delete all indexes under path
     */
    private void evictVirtualCache(String virtualKey, String metadataPath) {
        if (existLocally(virtualKey, metadataPath)) {
            log.debug("Evicting virtual cache at {}:{}", virtualKey, metadataPath);
            undeployArtifact(virtualKey, metadataPath);
        }
    }

    @VisibleForTesting
    List<PackageRepo> getVirtualReposContainingRepo(String repoKey)
            throws PackageException {
        RepoDescriptor repoDescriptor = repositoryService.repoDescriptorByKey(repoKey);
        if (repoDescriptor == null) {
            throw new PackageException(SC_NOT_FOUND, NOT_FOUND, "repo with name:" + repoKey + DOES_NOT_EXIST);
        }
        try {
            return repositoryService.getVirtualReposContainingRepo(repoDescriptor)
                    .stream()
                    .filter(virtualDescriptor -> repoDescriptor.getType().equals(virtualDescriptor.getType()))
                    .map(virtualDescriptor -> repositoryService.repositoryByKey(virtualDescriptor.getKey()))
                    .collect(Collectors.toList());
        } catch (Exception e) {
            throw new PackageException(SC_NOT_FOUND, NOT_FOUND, e.getMessage(), e);
        }
    }

}