package org.artifactory.projects;

import com.google.common.annotations.VisibleForTesting;
import lombok.NonNull;
import org.apache.commons.io.FileUtils;
import org.apache.commons.lang3.StringUtils;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.MutableStatusHolder;
import org.artifactory.projects.exception.AccessClientStateException;
import org.artifactory.projects.job.SyncRepositoriesWithAccessJob;
import org.artifactory.schedule.Task;
import org.artifactory.schedule.TaskService;
import org.artifactory.schedule.TaskUtils;
import org.artifactory.security.AuthenticationHelper;
import org.artifactory.security.SimpleUser;
import org.artifactory.security.UserGroupInfo;
import org.artifactory.security.access.AccessClientUpdateListener;
import org.artifactory.security.access.AccessService;
import org.artifactory.spring.Reloadable;
import org.artifactory.storage.StorageService;
import org.artifactory.util.CollectionUtils;
import org.artifactory.util.RepoPathUtils;
import org.jfrog.access.client.AccessClient;
import org.jfrog.access.client.AccessClientException;
import org.jfrog.access.client.group.GroupsClient;
import org.jfrog.access.client.project.GrpcProjectClient;
import org.jfrog.access.client.project.ProjectEntity;
import org.jfrog.access.client.project.ProjectResourceEntity;
import org.jfrog.access.client.project.ProjectSlimEntity;
import org.jfrog.access.client.project.RoleActions;
import org.jfrog.access.client.project.StorageQuotaEntity;
import org.jfrog.access.client.user.UsersClient;
import org.jfrog.access.common.ResourceType;
import org.jfrog.access.common.RoleConstants;
import org.jfrog.access.rest.user.User;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Primary;
import org.springframework.security.core.Authentication;
import org.springframework.stereotype.Service;

import javax.annotation.Nullable;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * @author BarakH
 */
@Service
@Primary
@Reloadable(beanClass = ProjectsService.class, initAfter = {AccessService.class, TaskService.class})
public class ProjectsServiceImpl implements ProjectsService, AccessClientUpdateListener {

    private static final Logger log = LoggerFactory.getLogger(ProjectsServiceImpl.class);
    private final AtomicBoolean isProjectStorageQuotaNotificationJobRunning = new AtomicBoolean(false);
    private final Object clientLock = new Object();
    private AccessService accessService;
    private GrpcProjectClient grpcProjectClient;
    private StorageService storageService;
    private TaskService taskService;

    private static SimpleUser getSimpleUser(@NonNull Authentication authentication) {
        return (SimpleUser) authentication.getPrincipal();
    }

    @Autowired
    public void setAccessService(AccessService accessService) {
        this.accessService = accessService;
    }

    @Autowired
    public void setStorageService(StorageService storageService) {
        this.storageService = storageService;
    }

    @Autowired
    public void setTaskService(TaskService taskService) {
        this.taskService = taskService;
    }

    @Override
    public void onContextCreated() {
        accessService.subscribeForClientChanges(this);
        startStorageQuotaNotificationJobIfNeeded();
        startSyncResourcesWithAccessJob();
    }

    @Override
    public void onClientUpdate() {
        synchronized (clientLock) {
            grpcProjectClient = null;
        }
    }

    @VisibleForTesting
    GrpcProjectClient getGrpcProjectClient() {
        synchronized (clientLock) {
            if (grpcProjectClient == null) {
                log.debug("Initiating new cachedGrpcProject client");
                AccessClient accessClient = accessService.getAccessClient();
                if (accessClient == null) {
                    throw new AccessClientStateException(
                            "Failed to acquire Access project gRPC client. Access might not be initiated");
                }
                this.grpcProjectClient = accessService.getAccessClient().cachedGrpcProject();
            }
            return this.grpcProjectClient;
        }
    }

    @Override
    public boolean isProjectAdmin(@Nullable String projectKey, @NonNull SimpleUser user) {
        if (ProjectsUtils.isDefaultProject(projectKey)) {
            return false;
        }
        log.debug("Checking if user '{}' with groups {} isProjectAdmin for projectKey '{}'",
                user.getUsername(), getUserGroupsNames(user), projectKey);
        try {
            return getGrpcProjectClient().isProjectAdmin(user.getUsername(), projectKey, getUserGroupsNames(user));
        } catch (AccessClientException e) {
            log.error("Failed to check if user {} is projectKey admin of projectKey {}: {}", user.getUsername(),
                    projectKey,
                    e.getMessage());
            log.debug("", e);
            return false;
        }
    }

    @Override
    public boolean isResourceManager(@NonNull SimpleUser user) {
        log.debug("Checking if user '{}' with groups {} isProjectAdmin for projectKey '{}'",
                user.getUsername(), getUserGroupsNames(user), ProjectsService.DEFAULT_PROJECT);
        return getGrpcProjectClient().isProjectAdmin(user.getUsername(), ProjectsService.DEFAULT_PROJECT,
                getUserGroupsNames(user));
    }

    @Override
    public void syncProjectResources(ResourceType resourceType, Set<String> resources) {
        log.debug("Sync resources of type: {} and values: {} with Access", resourceType,
                Arrays.toString(resources.toArray()));
        try {
            getGrpcProjectClient().syncResources(resourceType, resources);
        } catch (AccessClientException e) {
            log.error("Failed to sync resources of type: {} and values: {} with Access. Exception: {}",
                    resourceType, resources, e.getMessage());
            log.debug("", e);
        }
    }

    @Override
    public boolean canManageProjectResources(@Nullable String projectKey, @NonNull SimpleUser user) {
        if (ProjectsUtils.isDefaultProject(projectKey)) {
            return false;
        }
        log.debug("Checking if user '{}' with groups {} canManageProjectResources for projectKey '{}'",
                user.getUsername(), getUserGroupsNames(user), projectKey);
        try {
            return getGrpcProjectClient()
                    .canDoAction(user.getUsername(), projectKey, getUserGroupsNames(user),
                            RoleActions.MANAGE_RESOURCES);
        } catch (AccessClientException e) {
            log.error("Failed to check if user {} is can manage resources in projectKey {}: {}", user.getUsername(),
                    projectKey, e.getMessage());
            log.debug("", e);
            return false;
        }
    }

    @Override
    public boolean canProjectAdminManageResources(String repoKey, @NonNull SimpleUser user) {
        log.debug("Checking if user '{}' with groups {} can manage resource {}",
                user.getUsername(), getUserGroupsNames(user), repoKey);
        try {
            return getGrpcProjectClient()
                    .canProjectAdminDoOnResource(repoKey, ResourceType.REPO, List.of(RoleActions.MANAGE_RESOURCES),
                            user.getUsername(), getUserGroupsNames(user));
        } catch (AccessClientException e) {
            log.error("Failed to check if user {} is can manage resources {}: {}", user.getUsername(),
                    repoKey, e.getMessage());
            log.debug("", e);
            return false;
        }
    }

    @Override
    public void addProjectRepository(@NonNull String repoKey, String projectKey, Set<String> environments) {
        if (ProjectsUtils.isDefaultProject(projectKey)) {
            return;
        }
        try {
            log.debug("Adding repoKey '{}' into projectKey '{}' with environments '{}'", repoKey, projectKey,
                    environments);
            getGrpcProjectClient().addProjectResource(projectKey, ResourceType.REPO, repoKey, environments, true);
            log.debug("RepoKey '{}' was added into projectKey '{}' with environments '{}'", repoKey, projectKey,
                    environments);
        } catch (AccessClientException e) {
            log.error("Failed to add repository {} into projectKey {} and environment {}. {}", repoKey, projectKey,
                    environments.toArray(), e.getMessage());
            log.debug("", e);
            throw e;
        }
    }

    @Override
    public void updateProjectRepository(@NonNull String repoKey, Set<String> environments) {
        String projectKey = getProjectKeyOfRepository(repoKey);
        if (ProjectsUtils.isDefaultProject(projectKey)) {
            log.debug("Skipping on updating repoKey {} environments for project {}", repoKey, projectKey);
            return;
        }
        log.debug("Updating repoKey '{}' from project '{}' with environments '{}'", repoKey, projectKey, environments);
        try {
            getGrpcProjectClient().addProjectResource(projectKey, ResourceType.REPO, repoKey, environments, false);
        } catch (AccessClientException e) {
            log.error("Failed to update repository {} in projectKey {} with environment {}. {}", repoKey, projectKey,
                    environments.toArray(), e.getMessage());
            log.debug("", e);
            throw e;
        }
    }

    @Override
    @NonNull
    public List<String> getAllProjectsManagedByUser() {
        log.debug("Trying to get all the projectKeys the current user has projectAdmin role on");
        return getUserProjectsFromAccess();
    }

    @Override
    public @NonNull List<String> getAllProjects() {
        log.debug("Trying to retrieve all the projectKeys");
        try {
            return getGrpcProjectClient().getAllProjectKeys();
        } catch (Exception e) {
            log.error("Failed to get all the projects keys from Access with an error {}", e.getMessage());
            log.debug("", e);
        }
        return Collections.emptyList();
    }

    @Override
    public @NonNull List<ProjectEntity> getAllProjectsEntities() {
        log.debug("Trying to retrieve to retrieve all the ProjectsEntities");
        try {
            return getGrpcProjectClient().getAllProjects();
        } catch (Exception e) {
            log.error("Failed to get all the projects entities from Access with an error {}", e.getMessage());
            log.debug("", e);
        }
        return List.of();
    }

    @Override
    public void removeRepositoryFromCurrentProject(@NonNull String repoKey) {
        String currentProject = getProjectKeyOfRepository(repoKey);
        try {
            log.debug("Deleting repository '{}' from project: '{}'", repoKey, currentProject);
            getGrpcProjectClient().removeProjectResource(currentProject, ResourceType.REPO, repoKey);
            log.debug("Deleted repository '{}' from project: '{}'", repoKey, currentProject);
        } catch (AccessClientException e) {
            log.error("Failed to remove repository {} from projectKey {}. {}", repoKey, currentProject, e.getMessage());
            log.debug("", e);
            throw e;
        }
    }

    @Override
    public @NonNull Set<String> getProjectRepositoriesKeys(@NonNull String projectKey, boolean ignoreCache) {
        try {
            log.debug(
                    "Trying to retrieve all the assigned and shared repositories of project {} with ignoreCache set " +
                            "to {}", projectKey, ignoreCache);
            return getGrpcProjectClient().getProjectResourceEntitiesMap(projectKey, ResourceType.REPO, ignoreCache)
                    .keySet();
        } catch (Exception e) {
            log.error("Failed to find projectKey {} resources in Access: {}", projectKey, e.getMessage());
            log.debug("", e);
        }
        return Collections.emptySet();
    }

    @Override
    @NonNull
    public String getProjectKeyOfRepository(@Nullable String repoKey) {
        return getProjectKeyOfRepo(repoKey);
    }

    @NonNull
    private String getProjectKeyOfRepo(@Nullable String repoKey) {
        log.debug("Trying to retrieve the assigned project of repoKey {}", repoKey);
        if (StringUtils.isEmpty(repoKey)) {
            return DEFAULT_PROJECT;
        }
        repoKey = RepoPathUtils.removeDashCache(repoKey);
        return getResourceProjectFromAccess(repoKey, ResourceType.REPO);
    }

    @Override
    @NonNull
    public String getProjectDisplayName(String projectKey) {
        if (ProjectsUtils.isDefaultProject(projectKey)) {
            return DEFAULT_PROJECT;
        }
        try {
            log.debug("Trying to retrieve project display name for projectKey {}", projectKey);
            return getGrpcProjectClient().getProjectDisplayName(projectKey);
        } catch (Exception e) {
            log.error("Failed to find project display name for projectKey {} in Access: {}", projectKey,
                    e.getMessage());
            log.debug("", e);
        }
        return DEFAULT_PROJECT;
    }

    @Override
    @NonNull
    public Map<String, String> getRepoKeyToProjectMap(@NonNull Set<String> repositories, boolean ignoreCache) {
        repositories = repositories.stream()
                .map(RepoPathUtils::removeDashCache)
                .collect(Collectors.toSet());
        try {
            log.debug(
                    "Trying to retrieve projects assignments for the following repositories {} with ignoreCache set " +
                            "to {}", repositories, ignoreCache);
            return getGrpcProjectClient().getProjectForResourceMap(repositories, ResourceType.REPO, ignoreCache);
        } catch (Exception e) {
            log.error("Failed to find projects for the following repositories {} in Access: {}", repositories.toArray(),
                    e.getMessage());
            log.debug("", e);
        }
        return Map.of();
    }

    @Override
    public void validateRepoConfigIsMatchingToProjectSpec(@NonNull String projectKey, @NonNull String repoKey,
            Set<String> environments) {
        if (ProjectsUtils.isDefaultProject(projectKey)) {
            return;
        }
        log.debug("Verifying if repoKey {} with environments {} is a valid convention for project {}",
                repoKey, environments, projectKey);
        ProjectSlimEntity projectEntity = getGrpcProjectClient().getProjectSlimEntity(projectKey);
        if (projectEntity == null) {
            throw new IllegalArgumentException("Project " + projectKey + " does not exist");
        }
        if (!repoKey.startsWith(projectEntity.getProjectKey().concat("-"))) {
            throw new IllegalArgumentException(
                    "Repository key: " + repoKey + ", should start with the projectKey and a dash: '" +
                            projectEntity.getProjectKey().concat("-'"));
        }
        if (CollectionUtils.isNullOrEmpty(environments)) {
            throw new IllegalArgumentException(
                    "You must include at least one environment in the repository configuration");
        }
        //At this point there are only 2 global environments which are not unique to each projectKey - it might change
        //in the future
        if (!GLOBAL_ENVIRONMENTS.containsAll(environments)) {
            throw new IllegalArgumentException(
                    "Cannot set environments that are not part of the project. " +
                            "Project environments: [" + ENV_PROD + ", " + ENV_DEV + "] but got: " + environments);
        }
    }

    @Override
    public long getProjectStorageQuotaBytes(String projectKey) {
        log.debug("Trying to retrieve the project storage quota of project {}", projectKey);
        if (ProjectsUtils.isDefaultProject(projectKey)) {
            return UNLIMITED_STORAGE;
        }
        ProjectSlimEntity projectEntity = getProjectSpec(projectKey);
        if (projectEntity == null || projectEntity.getStorageQuota() == null) {
            return UNLIMITED_STORAGE;
        }
        Long storageQuotaBytes = projectEntity.getStorageQuota().getStorageQuotaBytes();
        log.debug("Storage quota of project {} is set to {}", projectKey,
                storageQuotaBytes);
        return storageQuotaBytes;
    }

    @Nullable
    @Override
    public ProjectResourceEntity getProjectRepositoryEntity(@NonNull String projectKey, @NonNull String repoKey) {
        try {
            log.debug("Trying to retrieve ProjectResourceEntity of repoKey {} from project {}", repoKey, projectKey);
            return getProjectRepositoriesMap(projectKey, false).getOrDefault(repoKey, null);
        } catch (Exception e) {
            log.error("Failed to find projectKey {} resources in Access: {}", projectKey, e.getMessage());
            log.debug("", e);
        }
        return null;
    }

    @Nullable
    @Override
    public ProjectSlimEntity getProjectSpec(String projectKey) {
        try {
            log.debug("Trying to retrieve ProjectSlimEntity of project {}", projectKey);
            return getGrpcProjectClient().getProjectSlimEntity(projectKey);
        } catch (Exception e) {
            log.error("Failed to get projectKey {} entity from Access.", projectKey, e);
        }
        return null;
    }

    @Override
    @NonNull
    public Map<String, Map<String, ProjectResourceEntity>> getAllProjectsRepositoriesMap(boolean ignoreCache) {
        log.debug("Trying to retrieve all projects and repositories map with ignoreCache set to {}", ignoreCache);
        return getAllProjects().stream().collect(Collectors.toMap(String::toString,
                project -> getProjectRepositoriesMap(project, ignoreCache)));
    }

    @Override
    @NonNull
    public Map<String, ProjectResourceEntity> getProjectRepositoriesMap(@NonNull String projectKey,
            boolean ignoreCache) {
        try {
            log.debug("Trying to retrieve project resource entities map of project {} with ignoreCache set to {}",
                    projectKey, ignoreCache);
            return getGrpcProjectClient().getProjectResourceEntitiesMap(projectKey, ResourceType.REPO, ignoreCache);
        } catch (Exception e) {
            log.error("Failed to find project {} resources in Access: {}", projectKey, e.getMessage());
            log.debug("", e);
        }
        return Map.of();
    }

    @Override
    public void assertProjectStorageQuota(String repoKey, MutableStatusHolder statusHolder, long contentLength) {
        long projectLimit = getProjectQuotaSizeByRepository(repoKey);
        if (projectLimit == UNLIMITED_STORAGE) {
            return;
        }
        String project = getProjectKeyOfRepository(repoKey);
        long projectTotalSize = storageService.getProjectStorageUsage(project);
        log.debug("Project {} has total storage usage of {} bytes out of {} bytes limit.", project, projectTotalSize,
                projectLimit);

        long expectedProjectTotalSize = projectTotalSize + contentLength;
        if (expectedProjectTotalSize > projectLimit) {
            ProjectSlimEntity projectEntity = Objects.requireNonNull(getProjectSpec(project));
            boolean isSoftLimit =
                    Optional.ofNullable(projectEntity.getStorageQuota()).map(StorageQuotaEntity::isSoftLimit)
                            .orElse(false);
            ProjectsUtils.handleStatusUpdateByQuotaLimitType(statusHolder, isSoftLimit, log);
            printQuotaExceededOnDebug(projectLimit, project, expectedProjectTotalSize, isSoftLimit);
        }
    }

    private void printQuotaExceededOnDebug(long projectLimit, String project, long expectedProjectTotalSize,
            boolean isSoftLimit) {
        String limitType = isSoftLimit ? "soft" : "hard";
        if (log.isDebugEnabled()) {
            log.debug("Project {} is about to use {} of storage which exceeds the {} limit of {}.",
                    project,
                    FileUtils.byteCountToDisplaySize(expectedProjectTotalSize),
                    limitType,
                    FileUtils.byteCountToDisplaySize(projectLimit));
        }
    }

    @Override
    public long getProjectQuotaSizeByRepository(@NonNull String repoKey) {
        if (!storageService.isProjectsStorageQuotaEnabled()) {
            log.debug("Projects storage quota is disabled.");
            return ProjectsService.UNLIMITED_STORAGE;
        }
        String project = getProjectKeyOfRepository(repoKey);
        if (ProjectsUtils.isDefaultProject(project)) {
            log.debug("Project is either empty or default for repository key: '{}'", repoKey);
            return ProjectsService.UNLIMITED_STORAGE;
        }
        long projectStorageQuotaBytes = getProjectStorageQuotaBytes(project);
        log.debug("Project storage quota is: '{}'", projectStorageQuotaBytes);
        return projectStorageQuotaBytes;
    }

    private void startStorageQuotaNotificationJobIfNeeded() {
        if (isProjectStorageQuotaNotificationJobRunning.get()) {
            return;
        }
        Task task = TaskUtils.createRepeatingTask(ProjectStorageQuotaNotificationJob.class,
                TimeUnit.MINUTES.toMillis(ConstantValues.projectsStorageQuotaNotificationJobIntervalMinutes.getInt()),
                TimeUnit.MINUTES
                        .toMillis(ConstantValues.projectsStorageQuotaNotificationJobInitialDelayMinutes.getInt()),
                "Projects Storage Quota notification emails sender job");
        taskService.startTask(task, false);
        isProjectStorageQuotaNotificationJobRunning.set(true);
        log.info("Scheduling ProjectStorageQuotaNotificationJob to run every {} minutes",
                ConstantValues.projectsStorageQuotaNotificationJobIntervalMinutes.getInt());
    }

    @Override
    public @NonNull Set<User> getProjectAdmins(@NonNull ProjectEntity projectKey) {
        UsersClient usersClient = accessService.getAccessClient().users();
        GroupsClient groupsClient = accessService.getAccessClient().groups();

        Set<User> projectAdminUsers = projectKey.getUsers().entrySet().stream()
                .filter(userToRolesEntry -> userToRolesEntry.getValue().getRoles()
                        .contains(RoleConstants.PROJECT_ADMIN))
                .map(Map.Entry::getKey)
                .flatMap(username -> usersClient.findUserByUsername(username).stream())
                .collect(Collectors.toSet());

        Set<User> projectAdminGroupsUsers = projectKey.getGroups().entrySet().stream()
                .filter(groupToRolesEntry -> groupToRolesEntry.getValue().getRoles()
                        .contains(RoleConstants.PROJECT_ADMIN))
                .map(Map.Entry::getKey)
                .flatMap(group -> groupsClient.findGroupUsers(group).getUsers().stream())
                .collect(Collectors.toSet());

        return Stream.of(projectAdminUsers, projectAdminGroupsUsers)
                .flatMap(Collection::stream)
                .collect(Collectors.toSet());
    }

    @NonNull
    private List<String> getUserProjectsFromAccess() {
        Authentication authentication = AuthenticationHelper.getAuthentication();
        if (authentication == null) {
            log.debug("User is not authenticated");
            return Collections.emptyList();
        }
        SimpleUser user = getSimpleUser(authentication);
        try {
            if (user.isAdmin()) {
                log.debug("User {} is super admin, returning all the projectKeys", user.getUsername());
                return getGrpcProjectClient().getAllProjectKeys();
            }
            List<String> groups = getUserGroupsNames(user);
            log.debug("Trying to retrieve all the projectKeys that user {} with groups {} has project admin role on",
                    user.getUsername(), groups);
            return getGrpcProjectClient().getAllProjectsManagedByUser(user.getUsername(), groups)
                    .stream()
                    .filter(projectKey -> !DEFAULT_PROJECT.equals(projectKey))
                    .collect(Collectors.toList());
        } catch (Exception e) {
            log.error("Failed to find user {} projects in access. {}", user.getUsername(), e.getMessage());
            log.debug("", e);
        }
        return Collections.emptyList();
    }

    @NonNull
    private String getResourceProjectFromAccess(@NonNull String resourceName, ResourceType resourceType) {
        try {
            log.debug("Trying to retrieve the assigned project key of resource name '{}' from resource type {}",
                    resourceName, resourceType);
            return getGrpcProjectClient().getProjectForResource(resourceName, resourceType);

        } catch (AccessClientStateException e) {
            log.warn("Failed to find the project key of '{}' from {} resource type in access: {}", resourceName,
                    resourceType.getKey(), e.getMessage());
        } catch (Exception e) {
            log.error("Failed to find the project key of '{}' from {} resource type in access: {}", resourceName,
                    resourceType.getKey(), e.getMessage());
            log.debug("", e);
        }
        return DEFAULT_PROJECT;
    }

    @Nullable
    private List<String> getUserGroupsNames(SimpleUser simpleUser) {
        Set<UserGroupInfo> userGroups = simpleUser.getDescriptor().getGroups();
        if (userGroups.isEmpty()) {
            return null;
        }
        return userGroups.stream().map(UserGroupInfo::getGroupName).collect(Collectors.toList());
    }

    private void startSyncResourcesWithAccessJob() {
        Task deleteReposFromAccessTask =
                TaskUtils.createRepeatingTask(SyncRepositoriesWithAccessJob.class,
                        ConstantValues.syncResourcesWithAccessJobInterval.getLong(), 0, "Delete repos from Access");
        taskService.startTask(deleteReposFromAccessTask, false);
    }
}
