package org.artifactory.servicetrust;

import org.artifactory.api.security.access.GenericTokenSpec;
import org.artifactory.api.security.access.TokenSpec;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.jfrog.access.common.ServiceId;
import org.jfrog.common.JsonUtils;

import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import static org.artifactory.servicetrust.ServiceTrustService.DEFAULT_EXCHANGE_TOKEN_EXPIRATION_TIME_IN_SECS;
import static org.artifactory.servicetrust.ServiceTrustService.DEFAULT_MASTER_TOKEN_EXPIRATION_TIME_IN_SECS;
import static org.artifactory.servicetrust.ServiceTrustService.DEFAULT_PAIRING_TOKEN_EXPIRATION_TIME_IN_SECS;
import static org.artifactory.servicetrust.ServiceTrustService.INTERNAL_SCOPE_PREFIX;
import static org.artifactory.servicetrust.ServiceTrustService.INTERNAL_SERVICE_TRUST_PAIRING_SCOPE_POSTFIX;
import static org.artifactory.servicetrust.ServiceTrustService.INTERNAL_SERVICE_TRUST_PAIRING_SCOPE_PREFIX;
import static org.artifactory.servicetrust.ServiceTrustService.PAIRING_SCOPE_PATTERN;
import static org.jfrog.access.common.AccessAuthz.ADMIN;

public abstract class TokenSpecServiceBaseImpl implements TokenSpecOutboundService, TokenSpecInboundService {

    @Override
    public TokenSpec<GenericTokenSpec> createPairingTokenSpec(@NotNull ServiceId artifactoryServiceId,
            @NotNull ServiceTrust trust, @NotNull String subject, @NotNull String scope, @NotNull String baseUrl,
            @Nullable String pairingUrl, @NotNull String exchangeUrl) {
        String useCasePath = getUseCasePathForScope(scope);
        String updatedScope = String.format(PAIRING_SCOPE_PATTERN, useCasePath);
        TokenSpec<GenericTokenSpec> spec = GenericTokenSpec.create(subject);
        spec.audience(Collections.singletonList(artifactoryServiceId.getFormattedName()));
        spec.expiresIn(DEFAULT_PAIRING_TOKEN_EXPIRATION_TIME_IN_SECS);
        spec.scope(Collections.singletonList(updatedScope));
        spec.refreshable(false);
        TokenExtension tokenExtension = new TokenExtension(baseUrl, exchangeUrl, pairingUrl, true);
        String extension = JsonUtils.getInstance().valueToString(tokenExtension);
        spec.extension(extension);
        return spec;
    }

    @NotNull
    protected abstract String getUseCasePathForScope(@NotNull String scope);

    protected abstract String getUseCaseFromPairScope(List<String> scopes);

    @Override
    public TokenSpec<GenericTokenSpec> createMasterTokenSpec(@NotNull ServiceId artifactoryServiceId,
            @NotNull String useCasePath, @NotNull String pairingScope, @NotNull String baseUrl,
            @NotNull String pairingUrl, @NotNull String exchangeUrl) {

        String subject = artifactoryServiceId.getFormattedName();
        TokenSpec<GenericTokenSpec> spec = GenericTokenSpec.create(subject);
        spec.audience(Collections.singletonList(artifactoryServiceId.getFormattedName()));
        spec.expiresIn(DEFAULT_MASTER_TOKEN_EXPIRATION_TIME_IN_SECS);
        String adminScope = artifactoryServiceId.getFormattedName() + ":" + ADMIN;
        String requiredScope = createRequiredInternalScope(useCasePath);
        spec.scope(Arrays.asList(adminScope, requiredScope, pairingScope));
        spec.refreshable(true);
        TokenExtension newExtension = new TokenExtension(baseUrl, exchangeUrl, pairingUrl, true);
        String extension = JsonUtils.getInstance().valueToString(newExtension);
        spec.extension(extension);
        return spec;
    }

    @Override
    @NotNull
    public String createRequiredInternalScope(@NotNull String useCasePath) {
        if(useCasePath.startsWith(INTERNAL_SCOPE_PREFIX))
            return useCasePath;

        return INTERNAL_SCOPE_PREFIX + useCasePath;
    }

    public TokenSpec<GenericTokenSpec> createExchangeTokenSpec(ServiceId artifactoryServiceId, String useCasePath) {
        String subject = artifactoryServiceId.getFormattedName();
        TokenSpec<GenericTokenSpec> spec = GenericTokenSpec.create(subject);
        spec.audience(List.of(artifactoryServiceId.getFormattedName()));
        spec.expiresIn(DEFAULT_EXCHANGE_TOKEN_EXPIRATION_TIME_IN_SECS);
        String adminScope = artifactoryServiceId.getFormattedName() + ":" + ADMIN;
        String requiredScope = createRequiredInternalScope(useCasePath);
        List<String> scopeList = List.of(adminScope, requiredScope);
        spec.scope(scopeList);
        spec.refreshable(false);
        String extension = JsonUtils.getInstance().valueToString(TokenExtension.REVOCABLE_TOKEN);
        spec.extension(extension);
        return spec;
    }

    @NotNull
    protected String getSimpleScope(String fullScope) {
        return fullScope
                .replace(INTERNAL_SERVICE_TRUST_PAIRING_SCOPE_PREFIX, "")
                .replace(INTERNAL_SERVICE_TRUST_PAIRING_SCOPE_POSTFIX, "");
    }
}
