package org.artifactory.maven;

import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.ArtifactoryRunningMode;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.maven.MavenMetadataService;
import org.artifactory.api.maven.MavenMetadataWorkItem;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RepoLayout;
import org.artifactory.descriptor.repo.RepoLayoutBuilder;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.model.common.RepoPathImpl;
import org.artifactory.repo.LocalRepo;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.db.DbLocalRepo;
import org.artifactory.repo.service.InternalRepositoryService;
import org.artifactory.spring.InternalArtifactoryContext;
import org.artifactory.util.RepoLayoutUtils;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.testng.annotations.AfterTest;
import org.testng.annotations.BeforeTest;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.clearInvocations;
import static org.mockito.Mockito.doReturn;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.verifyNoInteractions;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;

@Test
public class MavenMetadataServiceImplTest {

    @Mock
    private InternalArtifactoryContext artifactoryContext;
    @Mock
    private InternalRepositoryService repositoryService;
    @Mock
    private AddonsManager addonsManager;
    @Mock
    private MavenMetadataService mavenMetadataService;

    private MavenMetadataServiceImpl service;

    @BeforeTest
    void setup() {
        MockitoAnnotations.initMocks(this);
        ArtifactoryContextThreadBinder.bind(artifactoryContext);
        service = spy(new MavenMetadataServiceImpl());
        service.setAddonsManager(addonsManager);
        service.setRepoService(repositoryService);

        doReturn(mavenMetadataService).when(artifactoryContext).beanForType(MavenMetadataService.class);
    }

    @AfterTest
    void tearDown() {
        ArtifactoryContextThreadBinder.unbind();
    }

    @Test
    public void testCalculateMavenPluginMetadata() {
        // given
        RepoPath repoPath = new RepoPathImpl("local_repo",
                "org/codehaus/mojo/maven-buildnumber-plugin/0.9.4/maven-buildnumber-plugin-0.9.4.pom", false);
        MavenMetadataWorkItem workItem = new MavenMetadataWorkItem(repoPath, false, true);
        LocalRepoDescriptor localRepoDescriptor = new LocalRepoDescriptor();// mock(LocalRepoDescriptor.class);
        LocalRepo localRepo = new DbLocalRepo(localRepoDescriptor, repositoryService, null);
        localRepoDescriptor.setType(RepoType.Maven);

        when(addonsManager.getArtifactoryRunningMode()).thenReturn(ArtifactoryRunningMode.PRO);
        when(repositoryService.localOrFederatedRepositoryByKey(repoPath.getRepoKey())).thenReturn(localRepo);

        MavenMetadataCalculator mavenMetadataCalculator = mock(MavenMetadataCalculator.class);
        when(service.getMavenMetadataCalculator(workItem)).thenReturn(mavenMetadataCalculator);

        // when
        service.calculateMavenMetadata(workItem);

        // then
        verify(addonsManager).getArtifactoryRunningMode();
        verify(repositoryService).localOrFederatedRepositoryByKey(repoPath.getRepoKey());
        verify(service).getMavenMetadataCalculator(workItem);
        verify(mavenMetadataCalculator).calculate();
        verify(mavenMetadataService).calculateMavenPluginsMetadataAsync(any());

        // given
        clearInvocations(addonsManager, repositoryService, service, mavenMetadataCalculator, mavenMetadataService);
        workItem = new MavenMetadataWorkItem(repoPath, false, false);
        when(service.getMavenMetadataCalculator(workItem)).thenReturn(mavenMetadataCalculator);

        // when
        service.calculateMavenMetadata(workItem);

        // then
        verify(addonsManager).getArtifactoryRunningMode();
        verify(repositoryService).localOrFederatedRepositoryByKey(repoPath.getRepoKey());
        verify(service).getMavenMetadataCalculator(workItem);
        verify(mavenMetadataCalculator).calculate();
        verifyNoInteractions(mavenMetadataService);
    }

    @Test(dataProvider = "skipCalculationDataProvider")
    public void TestSkipCalculation(RepoType type, RepoLayout repoLayout, ArtifactoryRunningMode runningModel,
            boolean expected) {
        assertEquals(service.skipCalculation(type, repoLayout, runningModel), expected,
                "Layout = " + repoLayout.getName());
    }

    @DataProvider(name = "skipCalculationDataProvider")
    public Object[][] skipCalculationDataProvider() {
        return new Object[][] {
                // For backward comparability Maven indexing is supported on repositories with the right layout
                // even if they are not Maven family
                {RepoType.Generic, RepoLayoutUtils.MAVEN_2_DEFAULT, ArtifactoryRunningMode.OSS, false},
                {RepoType.Generic, RepoLayoutUtils.MAVEN_1_DEFAULT, ArtifactoryRunningMode.OSS, true},
                {RepoType.Gradle, RepoLayoutUtils.MAVEN_1_DEFAULT, ArtifactoryRunningMode.OSS, false},

                // Repos with null type should always be indexed - as used to be in the old code.
                {null, RepoLayoutUtils.MAVEN_2_DEFAULT, ArtifactoryRunningMode.PRO, false},
                {null, RepoLayoutUtils.MAVEN_1_DEFAULT, ArtifactoryRunningMode.PRO, false},
                {null, RepoLayoutUtils.MAVEN_2_DEFAULT, ArtifactoryRunningMode.JCR, false},

                // Maven family should never be skipped - even at wrong layout
                {RepoType.Ivy,
                        new RepoLayoutBuilder()
                                .name("FakeLayout").build(),
                        ArtifactoryRunningMode.PRO, false},

                // Maven 2 layout should be ignored if the license is not right and the repo is not Maven family
                {RepoType.Generic, RepoLayoutUtils.MAVEN_2_DEFAULT, ArtifactoryRunningMode.JCR, true},
                {RepoType.Generic, RepoLayoutUtils.MAVEN_2_DEFAULT, ArtifactoryRunningMode.AOL_JCR, true},
                {RepoType.Generic, RepoLayoutUtils.MAVEN_2_DEFAULT, ArtifactoryRunningMode.CONAN, true},

                // Maven family should never be skipped - even at wrong license
                {RepoType.Maven, RepoLayoutUtils.MAVEN_2_DEFAULT, ArtifactoryRunningMode.JCR, false}
        };
    }
}
