package org.artifactory.ph;

import com.google.common.collect.ArrayListMultimap;
import com.google.common.collect.Lists;
import com.google.common.collect.Multimap;
import org.artifactory.api.properties.PropertiesService;
import org.artifactory.common.repo.PackageType;
import org.artifactory.common.repo.RepoInfo;
import org.artifactory.common.repo.RepoType;
import org.artifactory.core.api.ph.exception.PackageException;
import org.artifactory.core.api.ph.virtual.resolution.PackageRepo;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RemoteRepoDescriptor;
import org.artifactory.descriptor.repo.RepoDescriptor;
import org.artifactory.descriptor.repo.VirtualRepoDescriptor;
import org.artifactory.model.xstream.fs.PropertiesImpl;
import org.artifactory.repo.LocalRepo;
import org.artifactory.repo.RealRepo;
import org.artifactory.repo.RemoteRepo;
import org.artifactory.repo.Repo;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.RepoPathFactory;
import org.artifactory.repo.service.InternalRepositoryService;
import org.artifactory.repo.virtual.VirtualRepo;
import org.artifactory.repo.virtual.VirtualRepositoriesResolver;
import org.artifactory.util.DoesNotExistException;
import org.junit.jupiter.api.Assertions;
import org.mockito.ArgumentCaptor;
import org.mockito.Mock;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import java.util.ArrayList;
import java.util.List;

import static org.apache.http.HttpStatus.SC_INTERNAL_SERVER_ERROR;
import static org.apache.http.HttpStatus.SC_NOT_FOUND;
import static org.artifactory.common.repo.RepoType.LOCAL;
import static org.artifactory.common.repo.RepoType.REMOTE;
import static org.artifactory.common.repo.RepoType.VIRTUAL;
import static org.artifactory.descriptor.repo.RepoType.CRAN;
import static org.artifactory.descriptor.repo.RepoType.Docker;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyBoolean;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.reset;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.verifyNoMoreInteractions;
import static org.mockito.Mockito.when;
import static org.mockito.MockitoAnnotations.openMocks;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertFalse;
import static org.testng.Assert.assertNotNull;
import static org.testng.Assert.assertNull;
import static org.testng.Assert.assertTrue;
import static org.testng.Assert.fail;

@SuppressWarnings("rawtypes")
@Test
public class PackageRepoServiceImplTest {

    @Mock
    private InternalRepositoryService repositoryService;
    @Mock
    private PropertiesService propertiesService;

    private PackageRepoServiceImpl packageRepoService;

    @BeforeMethod
    private void setUp() {
        openMocks(this);
        packageRepoService = new PackageRepoServiceImpl(repositoryService, propertiesService);
    }

    @Test
    public void testIsVirtual() {
        when(repositoryService.virtualRepositoryByKey(any())).thenReturn(null);
        assertFalse(packageRepoService.isVirtual("repoKey"));
        verify(repositoryService, times(1)).virtualRepositoryByKey(eq("repoKey"));

        reset(repositoryService);

        when(repositoryService.virtualRepositoryByKey(any())).thenReturn(mock(VirtualRepo.class));
        assertTrue(packageRepoService.isVirtual("repoKey"));
        verify(repositoryService, times(1)).virtualRepositoryByKey(eq("repoKey"));
    }

    @Test
    public void testIsRemote() {
        when(repositoryService.remoteRepositoryByKey(any())).thenReturn(null);
        assertFalse(packageRepoService.isRemote("repoKey"));
        verify(repositoryService, times(1)).remoteRepositoryByKey(eq("repoKey"));

        reset(repositoryService);

        when(repositoryService.remoteRepositoryByKey(any())).thenReturn(mock(RemoteRepo.class));
        assertTrue(packageRepoService.isRemote("repoKey"));
        verify(repositoryService, times(1)).remoteRepositoryByKey(eq("repoKey"));
    }

    @Test
    public void testIsLocal() {
        when(repositoryService.localOrFederatedRepositoryByKey(any())).thenReturn(null);
        assertFalse(packageRepoService.isLocal("repoKey"));
        verify(repositoryService, times(1)).localOrFederatedRepositoryByKey(eq("repoKey"));

        reset(repositoryService);

        when(repositoryService.localOrFederatedRepositoryByKey(any())).thenReturn(mock(LocalRepo.class));
        assertTrue(packageRepoService.isLocal("repoKey"));
        verify(repositoryService, times(1)).localOrFederatedRepositoryByKey(eq("repoKey"));
    }

    @Test
    public void testMove() throws PackageException {
        when(repositoryService.move(any(RepoPath.class), any(RepoPath.class), anyBoolean(), anyBoolean(),
                anyBoolean())).thenReturn(null);

        packageRepoService.move("repoKey", "repoKey", "source", "target");

        ArgumentCaptor<RepoPath> source = ArgumentCaptor.forClass(RepoPath.class);
        ArgumentCaptor<RepoPath> target = ArgumentCaptor.forClass(RepoPath.class);

        verify(repositoryService, times(1)).move(source.capture(), target.capture(), eq(false), eq(true), eq(true));

        assertEquals(source.getValue().toPath(), "repoKey/source");
        assertEquals(target.getValue().toPath(), "repoKey/target");

        reset(repositoryService);

        when(repositoryService.move(any(RepoPath.class), any(RepoPath.class), anyBoolean(), anyBoolean(),
                anyBoolean())).thenThrow(new RuntimeException("test"));

        try {
            packageRepoService.move("repoKey", "repoKey", "source", "target");
            fail();
        } catch (PackageException e) {
            assertEquals(e.getStatus(), SC_INTERNAL_SERVER_ERROR);
        }
    }

    @DataProvider
    private Object[][] testGetRepoInfoDP() {
        return new Object[][] {
                {(Runnable) () -> {
                    LocalRepoDescriptor descriptor = mock(LocalRepoDescriptor.class);
                    when(descriptor.getType()).thenReturn(CRAN);
                    LocalRepo repo = mock(LocalRepo.class);
                    when(repo.isLocal()).thenReturn(true);
                    when(repo.getDescriptor()).thenReturn(descriptor);
                    when(repositoryService.repositoryByKey(anyString())).thenReturn(repo);
                    when(repositoryService.localOrFederatedRepositoryByKey(anyString())).thenReturn(repo);
                }, LOCAL},
                {(Runnable) () -> {
                    RemoteRepoDescriptor descriptor = mock(RemoteRepoDescriptor.class);
                    when(descriptor.getType()).thenReturn(CRAN);
                    RemoteRepo repo = mock(RemoteRepo.class);
                    when(repo.isReal()).thenReturn(true);
                    when(repo.getDescriptor()).thenReturn(descriptor);
                    when(repositoryService.repositoryByKey(anyString())).thenReturn(repo);
                    when(repositoryService.remoteRepositoryByKey(anyString())).thenReturn(repo);
                }, REMOTE},
                {(Runnable) () -> {
                    VirtualRepoDescriptor descriptor = mock(VirtualRepoDescriptor.class);
                    when(descriptor.getType()).thenReturn(CRAN);
                    VirtualRepo repo = mock(VirtualRepo.class);
                    when(repo.getDescriptor()).thenReturn(descriptor);
                    when(repositoryService.repositoryByKey(anyString())).thenReturn(repo);
                    when(repositoryService.virtualRepositoryByKey(anyString())).thenReturn(repo);
                }, VIRTUAL},

        };
    }

    @Test(dataProvider = "testGetRepoInfoDP")
    public void testGetRepoInfo(Runnable env, RepoType expectedRepoType) throws PackageException {
        env.run();
        RepoInfo repoInfo = packageRepoService.getRepoInfo("repoKey");
        assertNotNull(repoInfo);
        assertEquals(repoInfo.getRepoType(), expectedRepoType);
        assertEquals(repoInfo.getPackageType(), PackageType.CRAN);
        verify(repositoryService, times(1)).repositoryByKey("repoKey");
        verifyNoMoreInteractions(repositoryService);
    }

    public void testGetRepoInfoNotFound() {
        try {
            packageRepoService.getRepoInfo("repoKey");
            Assertions.fail();
        } catch (PackageException e) {
            Assertions.assertTrue(true);
        }
    }

    @Test(dataProvider = "testGetRepoInfoDP")
    public void testGetPackageType(Runnable env, RepoType expectedRepoType) {
        env.run();
        PackageType packageType = packageRepoService.getPackageType("repoKey");
        if (expectedRepoType != null) {
            assertEquals(packageType, PackageType.CRAN);
        } else {
            assertNull(packageType);
        }
    }

    @Test
    public void testSetAttributes() {
        PropertiesImpl properties = new PropertiesImpl();
        properties.put("k1", "v1");
        properties.put("k2", "v2");
        when(propertiesService.getProperties(any(RepoPath.class))).thenReturn(properties);
        Multimap<String, String> props = ArrayListMultimap.create();
        props.put("k1", "v11");
        props.put("k2", null);
        props.put("k3", "v3");

        packageRepoService.setAttributes("repoKey", "path", props);
        ArgumentCaptor<RepoPath> repoPath = ArgumentCaptor.forClass(RepoPath.class);
        verify(propertiesService, times(1)).setProperties(repoPath.capture(), eq(properties), eq(true));

        assertEquals(repoPath.getValue().toPath(), "repoKey/path");
        assertEquals(properties.size(), 3);
        assertEquals(properties.get("k1"), List.of("v11"));
        ArrayList<Object> list = Lists.newArrayList();
        list.add(null);
        assertEquals(properties.get("k2"), list);
        assertEquals(properties.get("k3"), List.of("v3"));
    }

    @Test
    public void testExist() {
        when(repositoryService.exists(any(RepoPath.class))).thenReturn(false);
        assertFalse(packageRepoService.existLocally("repoKey", "path"));
        verify(repositoryService, times(1)).exists(eq(RepoPathFactory.create("repoKey", "path")));

        reset(repositoryService);

        when(repositoryService.exists(any(RepoPath.class))).thenReturn(true);
        assertTrue(packageRepoService.existLocally("repoKey", "path"));
        verify(repositoryService, times(1)).exists(eq(RepoPathFactory.create("repoKey", "path")));
    }

    @Test(dataProvider = "testExistWithVirtualRepoProvider")
    public void testExistWithVirtualRepo(boolean exists) {
        String repoKey = "virtual";
        String path = "path";
        VirtualRepo virtualRepo = mock(VirtualRepo.class);
        when(repositoryService.virtualRepositoryByKey(repoKey)).thenReturn(virtualRepo);
        when(virtualRepo.itemExists(path)).thenReturn(exists);
        assertEquals(exists, packageRepoService.existLocally(repoKey, path));
        verify(virtualRepo, times(1)).itemExists(path);
        verify(repositoryService, times(1)).virtualRepositoryByKey(repoKey);
        verifyNoMoreInteractions(repositoryService, virtualRepo);
    }

    @Test
    public void testUndeployArtifact() {
        packageRepoService.undeployArtifact("repoKey", "path");
        verify(repositoryService, times(1)).undeploy(eq(RepoPathFactory.create("repoKey", "path")));
    }

    @Test
    public void testGetSubRepositories() throws PackageException {
        VirtualRepo virtualRepo = mock(VirtualRepo.class);
        VirtualRepositoriesResolver resolver = mock(VirtualRepositoriesResolver.class);
        RealRepo<?> subRepo = mock(RealRepo.class);
        when(subRepo.isCache()).thenReturn(true);
        when(virtualRepo.getVirtualRepositoriesResolver()).thenReturn(resolver);
        when(repositoryService.virtualRepositoryByKey(anyString())).thenReturn(virtualRepo);
        when(resolver.assembleSearchRepositoriesList(any(RepoPath.class), anyBoolean())).thenReturn(List.of(subRepo));

        List<? extends PackageRepo> subs = packageRepoService.getSubRepositories("repoKey", "", false);

        verify(repositoryService, times(1)).virtualRepositoryByKey(eq("repoKey"));
        verify(virtualRepo, times(1)).getVirtualRepositoriesResolver();
        verify(resolver, times(1)).assembleSearchRepositoriesList(eq(RepoPathFactory.create("repoKey", "")), eq(false));
        assertEquals(subs, List.of(subRepo));

        subs = packageRepoService.getSubRepositories("repoKey", "", true);
        assertEquals(subs.size(), 0);

        try {
            when(repositoryService.virtualRepositoryByKey(anyString())).thenReturn(null);
            packageRepoService.getSubRepositories("repoKey", "", false);
            fail();
        } catch (PackageException e) {
            assertEquals(e.getStatus(), SC_NOT_FOUND);
        }
    }

    @Test
    public void testGetDefaultDeploymentRepoKey() throws PackageException {
        when(repositoryService.isVirtualRepoExist(anyString())).thenReturn(true);
        when(repositoryService.getDefaultDeploymentRepoKey(anyString())).thenReturn("depRepoKey");
        assertEquals(packageRepoService.getDefaultDeploymentRepoKey("repoKey"), "depRepoKey");
        verify(repositoryService, times(1)).getDefaultDeploymentRepoKey(eq("repoKey"));

        reset(repositoryService);

        try {
            when(repositoryService.isVirtualRepoExist(anyString())).thenReturn(true);
            when(repositoryService.getDefaultDeploymentRepoKey(anyString())).thenThrow(
                    new DoesNotExistException("error"));
            packageRepoService.getDefaultDeploymentRepoKey("repoKey");
            fail();
        } catch (PackageException e) {
            verify(repositoryService, times(1)).getDefaultDeploymentRepoKey(eq("repoKey"));
            assertEquals(e.getStatus(), SC_NOT_FOUND);
        }
    }

    @Test
    public void testGetRemoteRepoBaseUrl() throws PackageException {
        final String remoteKey = "repoKey";
        final String expectedUrl = "http:://domain.com";
        RemoteRepo remoteRepo = mock(RemoteRepo.class);
        when(remoteRepo.getUrl()).thenReturn(expectedUrl);
        when(repositoryService.remoteRepositoryByKey(remoteKey)).thenReturn(remoteRepo);
        String baseUrl = packageRepoService.getRemoteRepoBaseUrl("repoKey");
        assertEquals(expectedUrl, baseUrl);
    }

    @Test
    public void testGetVirtualReposContainingRepoByPackageType() throws PackageException {

        try {
            when(repositoryService.repoDescriptorByKey(anyString())).thenReturn(null);
            packageRepoService.getVirtualReposContainingRepo("repoKey");
            fail();
        } catch (PackageException e) {
            assertEquals(e.getStatus(), SC_NOT_FOUND);
        }

        reset(repositoryService);

        RepoDescriptor descriptor = mock(RepoDescriptor.class);
        when(descriptor.getType()).thenReturn(CRAN);
        when(repositoryService.repoDescriptorByKey(anyString())).thenReturn(descriptor);

        VirtualRepoDescriptor virtualRepoDescriptor1 = mock(VirtualRepoDescriptor.class);
        when(virtualRepoDescriptor1.getKey()).thenReturn("repoKey1");
        when(virtualRepoDescriptor1.getType()).thenReturn(CRAN);

        VirtualRepoDescriptor virtualRepoDescriptor2 = mock(VirtualRepoDescriptor.class);
        when(virtualRepoDescriptor2.getKey()).thenReturn("repoKey2");
        when(virtualRepoDescriptor2.getType()).thenReturn(Docker);

        when(repositoryService.getVirtualReposContainingRepo(any(RepoDescriptor.class))).thenReturn(
                List.of(virtualRepoDescriptor1, virtualRepoDescriptor2));


        Repo result = mock(Repo.class);
        when(repositoryService.repositoryByKey(anyString())).thenReturn(result);

        List<PackageRepo> repos = packageRepoService.getVirtualReposContainingRepo("repoKey");

        assertEquals(repos, List.of(result));
        verify(repositoryService, times(1)).repoDescriptorByKey(eq("repoKey"));
        verify(repositoryService, times(1)).getVirtualReposContainingRepo(eq(descriptor));
        verify(repositoryService, times(1)).repositoryByKey(anyString());
    }

    @DataProvider
    private Object[][] testExistWithVirtualRepoProvider() {
        return new Object[][] {
                {true}, {false}
        };
    }
}