package org.artifactory.event;

import lombok.NonNull;

import javax.annotation.Nullable;
import java.util.Set;
import java.util.stream.Stream;

/**
 * Add a new type here when you need to use the EventLogHintManager API.
 * Be sure to use different prefix for each type.
 * prefix is saved as part of the hint value in the node_events table e,g: "ts1.value"
 * The prefix allow multiple types with same value coexist for same event, so developer should not worry about what
 * kind of value is permitted
 *
 * @author avishayh
 */
public enum EventHintType {

    FederatedReplication("fdr"),
    RemoveRepo("rmr"),
    RemovePackage("rmp"),
    AsyncThread("ast"), // The AsyncThread hint is added on all the async threads in Artifactory.
    PackageIndexExtraInfo("pie");

    private final String prefix;

    EventHintType(String prefix) {
        assert prefix.length() == 3;
        this.prefix = prefix + ".";
    }

    public String getPrefix() {
        return prefix;
    }

    @Nullable
    public static String getValue(@Nullable Set<String> hints, @NonNull EventHintType hintType) {
        if (hints == null) {
            return null;
        }
        String prefix = hintType.getPrefix();
        return hints.stream()
                .filter(value -> value != null && value.startsWith(prefix))
                .map(value -> value.substring(prefix.length()))
                .findAny().orElse(null);
    }

    public static boolean contains(@Nullable Set<String> hints, @NonNull EventHintType... hintTypes) {
        if (hints == null) {
            return false;
        }

        return Stream.of(hintTypes).anyMatch(hintType -> hints.stream()
                .anyMatch(value -> value != null && value.startsWith(hintType.getPrefix())));
    }
}
