/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.descriptor.config;

import com.google.common.base.Predicate;
import com.google.common.collect.Iterables;
import org.apache.commons.lang3.SerializationUtils;
import org.artifactory.descriptor.backup.BackupDescriptor;
import org.artifactory.descriptor.property.PropertySet;
import org.artifactory.descriptor.property.SystemPropertyDescriptor;
import org.artifactory.descriptor.replication.LocalReplicationDescriptor;
import org.artifactory.descriptor.replication.RemoteReplicationDescriptor;
import org.artifactory.descriptor.repo.FederatedRepoDescriptor;
import org.artifactory.descriptor.repo.HttpRepoDescriptor;
import org.artifactory.descriptor.repo.KeyPairDescriptor;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.ProxyDescriptor;
import org.artifactory.descriptor.repo.RealRepoDescriptor;
import org.artifactory.descriptor.repo.RemoteRepoDescriptor;
import org.artifactory.descriptor.repo.RepoLayout;
import org.artifactory.descriptor.repo.VirtualRepoDescriptor;
import org.artifactory.descriptor.security.SecurityDescriptor;
import org.artifactory.util.AlreadyExistsException;
import org.artifactory.util.RepoLayoutUtils;
import org.jetbrains.annotations.NotNull;
import org.jfrog.access.common.ServiceType;
import org.mockito.Mockito;
import org.testng.Assert;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;

import static org.mockito.ArgumentMatchers.any;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertFalse;
import static org.testng.Assert.assertNotNull;
import static org.testng.Assert.assertNull;
import static org.testng.Assert.assertTrue;

/**
 * Tests the MutableCentralConfigDescriptorImpl class.
 *
 * @author Yossi Shaul
 */
@Test
public class CentralConfigDescriptorImplTest {
    private CentralConfigDescriptorImpl cc;

    @BeforeMethod
    public void initCentralConfig() {
        cc = CentralConfigTestUtils.initCentralConfig();
    }

    public void defaultsTest() {
        CentralConfigDescriptorImpl cc = new CentralConfigDescriptorImpl();
        assertNotNull(cc.getLocalRepositoriesMap(), "Local repos map should not be null");
        assertNotNull(cc.getRemoteRepositoriesMap(), "Remote repos map should not be null");
        assertNotNull(cc.getFederatedRepositoriesMap(), "Federated repos map should not be null");
        assertNotNull(cc.getVirtualRepositoriesMap(), "Virtual repos map should not be null");
        assertNotNull(cc.getBackups(), "Backups list should not be null");
        assertNotNull(cc.getProxies(), "Proxies list should not be null");
        assertNotNull(cc.getPropertySets(), "Property sets list should not be null");
        assertNull(cc.getIndexer(), "Indexer should not be null");
        assertNotNull(cc.getSecurity(), "Security should not be null");
        assertNull(cc.getServerName(), "Server name should not be null");
        assertNotNull(cc.getDateFormat(), "Date format should not be null");
        assertTrue(cc.getFileUploadMaxSizeMb() > 50,
                "Default max file upload size should be bigger than 50mb");
        assertFalse(cc.isOfflineMode(), "Offline mode should be false by default");
        assertNotNull(cc.getRepoLayouts(), "Repo layouts list should not be null");
        assertNotNull(cc.getRemoteReplications(), "Remote replication list should not be null");
        assertNotNull(cc.getLocalReplications(), "Local replication list should not be null");
    }

    public void uniqueKeyExistence() {
        assertFalse(cc.isKeyAvailable("local1"));
        assertFalse(cc.isRepositoryExists("backup2"));
        assertFalse(cc.isRepositoryExists("proxy2"));

        assertTrue(cc.isKeyAvailable("proxy22"));
        cc.setSecurity(new SecurityDescriptor());
        assertTrue(cc.isKeyAvailable("ldap1"));

        assertTrue(cc.isKeyAvailable("repo")); // Global virtual repo has been deprecated

        assertTrue(cc.isPropertySetExists("set2"));

        assertFalse(cc.isKeyAvailable("layout1"));
        assertTrue(cc.isRepoLayoutExists("layout2"));

        assertFalse(cc.isKeyAvailable("keyPair1"));
        assertFalse(cc.isKeyPairExists("keypair2"));
    }

    public void keyPairsExistence() {
        assertEquals(cc.getKeyPairs().size(), 1, "Key pairs count mismatch");
        assertTrue(cc.isKeyPairExists("keyPair1"));
        KeyPairDescriptor keypair1 = cc.getKeyPair("keyPair1");
        assertNotNull(keypair1);
        assertEquals(keypair1.getKey(), "keyPair1");
    }

    public void repositoriesExistence() {
        assertEquals(cc.getLocalRepositoriesMap().size(), 2, "Local repos count mismatch");
        assertEquals(cc.getRemoteRepositoriesMap().size(), 1, "Remote repos count mismatch");
        assertEquals(cc.getVirtualRepositoriesMap().size(), 1,
                "Virtual repos count mismatch");
        assertTrue(cc.isRepositoryExists("local1"));
        assertTrue(cc.isRepositoryExists("local2"));
        assertTrue(cc.isRepositoryExists("remote1"));
        assertTrue(cc.isRepositoryExists("federated1"));
        assertTrue(cc.isRepositoryExists("virtual1"));
    }

    public void proxyExistence() {
        assertEquals(cc.getProxies().size(), 2, "Proxies count mismatch");
        assertTrue(cc.isProxyExists("proxy1"));
        assertTrue(cc.isProxyExists("proxy2"));
    }

    public void reverseProxyKeyPrevention() {
        assertFalse(cc.isKeyAvailable("direct"));
        assertFalse(cc.isKeyAvailable("apache"));
        assertFalse(cc.isKeyAvailable("nginx"));
    }

    public void removeProxy() {
        assertTrue(cc.isProxyExists("proxy2"));
        ProxyDescriptor removedProxy = cc.removeProxy("proxy2");
        assertEquals(removedProxy.getKey(), "proxy2");
        assertEquals(cc.getProxies().size(), 1, "Only one proxy expected");
        assertFalse(cc.isProxyExists("proxy2"));
    }

    @Test(dataProvider = "proxyKeyProvider")
    public void addProxy(String proxyKey, boolean isValid) {
        ProxyDescriptor proxyDescriptor = new ProxyDescriptor();
        proxyDescriptor.setKey(proxyKey);
        try {
            cc.addProxy(proxyDescriptor, true);
            cc.removeProxy(proxyKey);
        } catch (IllegalArgumentException e) {
            if (!isValid) {
                Assert.assertNotNull(e);
            } else {
                Assert.fail("Proxy Key " + proxyDescriptor.getKey() + " should be valid");
            }
        }
    }

    public void removeProxyFromExistingRepo() {
        ProxyDescriptor proxyDescriptor = new ProxyDescriptor();
        proxyDescriptor.setKey("proxy3");
        cc.addProxy(proxyDescriptor, true);
        addDefaultProxyToFederatedRepositories(proxyDescriptor);
        cc.removeProxy("proxy3");
        assertNull(cc.getFederatedRepositoriesMap().get("federated1").getProxy());
    }

    public void removeReferencedProxy() {
        ProxyDescriptor proxy = new ProxyDescriptor();
        proxy.setKey("referencedProxy");
        cc.addProxy(proxy, true);

        HttpRepoDescriptor remoteRepo =
                (HttpRepoDescriptor) cc.getRemoteRepositoriesMap().get("remote1");
        remoteRepo.setProxy(proxy);

        assertNotNull(remoteRepo.getProxy(), "Just checking ...");

        LocalReplicationDescriptor localReplication = cc.getLocalReplication("local1");
        localReplication.setProxy(proxy);

        cc.removeProxy("referencedProxy");

        assertNull(remoteRepo.getProxy(), "Proxy should have been removed from the remote repo");
        assertNull(localReplication.getProxy(), "Proxy should have been removed from the local replication.");
    }

    public void handleDisableProxyForRepo() {
        // Check for empty proxy config
        HttpRepoDescriptor remoteRepo = (HttpRepoDescriptor) cc.getRemoteRepositoriesMap().get("remote1");
        assertNull(remoteRepo.getProxy(), "Proxy should be removed");
        assertFalse(remoteRepo.isDisableProxy(), "Proxy enabled");

        // Disable proxy for a single repo
        remoteRepo.setDisableProxy(true);
        remoteRepo.setProxy(null);
        cc.updateRemoteRepository(remoteRepo);

        // Create default proxy. Must update all repos with disableProxy=false.
        ProxyDescriptor proxy = new ProxyDescriptor();
        proxy.setKey("defaultProxy");
        proxy.setPlatformDefault(true);
        cc.addProxy(proxy, true);

        // Check all repo after added default proxy
        remoteRepo = (HttpRepoDescriptor) cc.getRemoteRepositoriesMap().get("remote1");
        assertNull(remoteRepo.getProxy(), "Remote proxy should be removed");
        assertTrue(remoteRepo.isDisableProxy(), "Remote proxy disabled");

        LocalReplicationDescriptor localReplication = cc.getLocalReplication("local1");
        assertNotNull(localReplication.getProxy(), "Local replication should have a proxy");
        assertEquals(localReplication.getProxy().getKey(), "defaultProxy", "Proxy name is default");

        FederatedRepoDescriptor federatedRepoDescriptor = cc.getFederatedRepositoriesMap().get("federated1");
        assertNotNull(federatedRepoDescriptor.getProxy(), "Federated repo should have a proxy");
        assertEquals(federatedRepoDescriptor.getProxy().getKey(), "defaultProxy", "Proxy name is default");

        // Enable proxy for a single repo
        remoteRepo.setDisableProxy(false);
        cc.updateRemoteRepository(remoteRepo);
        cc.proxyChanged(proxy, true);

        // Check proxy assign to default one
        remoteRepo = (HttpRepoDescriptor) cc.getRemoteRepositoriesMap().get("remote1");
        assertNotNull(remoteRepo.getProxy(), "Remote repo should have a proxy");
        assertEquals(remoteRepo.getProxy().getKey(), "defaultProxy", "Proxy name does not match");
    }

    public void platformProxyChanged() {
        ProxyDescriptor proxy = new ProxyDescriptor();
        proxy.setKey("defaultProxy");
        proxy.setPlatformDefault(true);
        cc.addProxy(proxy, true);

        ProxyDescriptor newDefaultProxy = new ProxyDescriptor();
        newDefaultProxy.setKey("newDefaultProxy");
        cc.addProxy(newDefaultProxy, true);
        newDefaultProxy.setPlatformDefault(true);
        // Change shouldn't update the proxies
        cc.proxyChanged(newDefaultProxy, false);

        HttpRepoDescriptor remoteRepo = (HttpRepoDescriptor) cc.getRemoteRepositoriesMap().get("remote1");
        assertNotNull(remoteRepo.getProxy(), "Remote repo should have a proxy");
        assertEquals(remoteRepo.getProxy().getKey(), "defaultProxy", "Proxy name does not match");

        LocalReplicationDescriptor localReplication = cc.getLocalReplication("local1");
        assertNotNull(localReplication.getProxy(), "Local replication should have a proxy");
        assertEquals(localReplication.getProxy().getKey(), "defaultProxy", "Proxy name does not match");

        FederatedRepoDescriptor federatedRepoDescriptor = cc.getFederatedRepositoriesMap().get("federated1");
        assertNotNull(federatedRepoDescriptor.getProxy(), "Federated repo should have a proxy");
        assertEquals(federatedRepoDescriptor.getProxy().getKey(), "defaultProxy", "Proxy name does not match");

        assertFalse(proxy.isPlatformDefault(), "Original proxy should now be false");

        // Change the platformdefault
        proxy.setPlatformDefault(true);
        newDefaultProxy.setPlatformDefault(true);
        // Change should update proxies
        cc.proxyChanged(newDefaultProxy, true);

        remoteRepo = (HttpRepoDescriptor) cc.getRemoteRepositoriesMap().get("remote1");
        assertNotNull(remoteRepo.getProxy(), "Remote repo should have a proxy");
        assertEquals(remoteRepo.getProxy().getKey(), "newDefaultProxy", "Proxy name does not match");

        localReplication = cc.getLocalReplication("local1");
        assertNotNull(localReplication.getProxy(), "Local replication should have a proxy");
        assertEquals(localReplication.getProxy().getKey(), "newDefaultProxy", "Proxy name does not match");

        federatedRepoDescriptor = cc.getFederatedRepositoriesMap().get("federated1");
        assertNotNull(federatedRepoDescriptor.getProxy(), "Federated repo should have a proxy");
        assertEquals(federatedRepoDescriptor.getProxy().getKey(), "newDefaultProxy", "Proxy name does not match");

        proxy.setPlatformDefault(false);
        cc.proxyChanged(proxy, true);

        assertNotNull(remoteRepo.getProxy(), "Remote repo should have a proxy");
        assertEquals(remoteRepo.getProxy().getKey(), "newDefaultProxy", "Proxy name does not match");

        assertNotNull(localReplication.getProxy(), "Remote repo should have a proxy");
        assertEquals(localReplication.getProxy().getKey(), "newDefaultProxy", "Proxy name does not match");

        assertNotNull(federatedRepoDescriptor.getProxy(), "Federated repo should have a proxy");
        assertEquals(federatedRepoDescriptor.getProxy().getKey(), "newDefaultProxy", "Proxy name does not match");
    }

    public void artifactoryProxyChanged() {
        ProxyDescriptor proxy = new ProxyDescriptor();
        proxy.setKey("defaultProxy");
        proxy.setServices(ServiceType.ARTIFACTORY);
        cc.addProxy(proxy, true);
        addDefaultProxyToRemoteRepositories(proxy);
        addDefaultProxyToFederatedRepositories(proxy);
        LocalReplicationDescriptor localReplication = cc.getLocalReplication("local1");
        localReplication.setProxy(proxy);

        ProxyDescriptor newArtifactoryProxy = new ProxyDescriptor();
        newArtifactoryProxy.setKey("newArtifactoryProxy");
        cc.addProxy(newArtifactoryProxy, true);
        newArtifactoryProxy.setServices(ServiceType.ARTIFACTORY);
        cc.proxyChanged(newArtifactoryProxy, true);
        HttpRepoDescriptor remoteRepo = (HttpRepoDescriptor) cc.getRemoteRepositoriesMap().get("remote1");

        assertNotNull(remoteRepo.getProxy(), "Remote repo should have a proxy");
        assertEquals(remoteRepo.getProxy().getKey(), "newArtifactoryProxy", "Proxy name does not match");

        assertNotNull(localReplication.getProxy(), "Local replication should have a proxy");
        assertEquals(localReplication.getProxy().getKey(), "newArtifactoryProxy", "Proxy name does not match");

        FederatedRepoDescriptor federatedRepoDescriptor = cc.getFederatedRepositoriesMap().get("federated1");
        assertNotNull(federatedRepoDescriptor.getProxy(), "Federated repo should have a proxy");
        assertEquals(federatedRepoDescriptor.getProxy().getKey(), "newArtifactoryProxy", "Proxy name does not match");

        assertFalse(proxy.isPlatformDefault(), "Original proxy should now be false");

        proxy.setServices(null);
        cc.proxyChanged(proxy, true);

        assertNotNull(remoteRepo.getProxy(), "Remote repo should have a proxy");
        assertEquals(remoteRepo.getProxy().getKey(), "newArtifactoryProxy", "Proxy name does not match");

        assertNotNull(localReplication.getProxy(), "Remote repo should have a proxy");
        assertEquals(localReplication.getProxy().getKey(), "newArtifactoryProxy", "Proxy name does not match");

        assertNotNull(federatedRepoDescriptor.getProxy(), "Federated repo should have a proxy");
        assertEquals(federatedRepoDescriptor.getProxy().getKey(), "newArtifactoryProxy", "Proxy name does not match");
    }

    public void proxyServicesChanged() {
        ProxyDescriptor proxy = new ProxyDescriptor();
        proxy.setKey("defaultProxy");
        proxy.setServices(String.join(",", ServiceType.ARTIFACTORY, ServiceType.XRAY));
        cc.addProxy(proxy, true);

        ProxyDescriptor newProxy = new ProxyDescriptor();
        newProxy.setKey("newProxy");
        cc.addProxy(newProxy, true);
        newProxy.setServices(ServiceType.ARTIFACTORY);
        cc.proxyChanged(newProxy, true);

        assertEquals(proxy.getServices(), ServiceType.XRAY, "Proxy service does not match");
        assertEquals(newProxy.getServices(), ServiceType.ARTIFACTORY, "Proxy service does not match");
        assertEquals(newProxy.getServicesList(), Collections.singletonList(ServiceType.ARTIFACTORY),
                "Proxy service does not match");

        assertFalse(proxy.isPlatformDefault(), "Original proxy should be false");

        //

        proxy.setPlatformDefault(true);
        proxy.setServices(null);
        cc.proxyChanged(proxy, true);

        assertNull(newProxy.getServices(), "Other proxy services should be null, as there is another default");
        assertFalse(newProxy.isPlatformDefault(), "New proxy should be false");

        // Testing removal of default proxy of OTHER proxy in list

        newProxy.setServices(ServiceType.ARTIFACTORY);
        cc.proxyChanged(newProxy, true);

        assertFalse(proxy.isPlatformDefault(), "Original proxy should be false");
        assertEquals(proxy.getServicesList().size(), 3, "The number of services should be all except Artifactory");
    }

    public void backupExistence() {
        assertEquals(cc.getBackups().size(), 3, "Backups count mismatch");
        assertTrue(cc.isBackupExists("backup1"));
        assertTrue(cc.isBackupExists("backup2"));
        assertTrue(cc.isBackupExists("backup3"));

    }

    public void removeBackup() {
        assertTrue(cc.isBackupExists("backup2"));
        BackupDescriptor removedBackup = cc.removeBackup("backup2");
        assertEquals(removedBackup.getKey(), "backup2");
        assertEquals(cc.getBackups().size(), 2, "Only 2 backups expected");
        assertFalse(cc.isBackupExists("backup2"));
    }

    public void excludeNewRepoFromBackup() {
        LocalRepoDescriptor newLocal = new LocalRepoDescriptor();
        newLocal.setKey("newLocal");
        cc.addLocalRepository(newLocal);
        cc.conditionallyAddToBackups(newLocal);
        RemoteRepoDescriptor newRemote = new HttpRepoDescriptor();
        newRemote.setKey("newRemote");
        newRemote.setUrl("https:localhost:8080/conan-smart-remote");
        cc.addRemoteRepository(newRemote);

        assertTrue(cc.isBackupExists("backup1"));
        BackupDescriptor backup1 = Iterables.tryFind(cc.getBackups(), new Predicate<BackupDescriptor>() {
            @Override
            public boolean apply(BackupDescriptor input) {
                return "backup1".equals(input.getKey());
            }
        }).orNull();
        assertNotNull(backup1, "Should find backup1");
        assertEquals(backup1.getExcludedRepositories().size(), 0);

        assertTrue(cc.isBackupExists("backup3"));
        BackupDescriptor backup3 = Iterables.tryFind(cc.getBackups(), new Predicate<BackupDescriptor>() {
            @Override
            public boolean apply(BackupDescriptor input) {
                return "backup3".equals(input.getKey());
            }
        }).orNull();
        assertNotNull(backup3, "Should find backup3");

        List<RealRepoDescriptor> excludedRepositories = backup3.getExcludedRepositories();
        assertEquals(excludedRepositories.size(), 2);
        assertTrue(excludedRepositories.contains(newLocal));
        assertTrue(excludedRepositories.contains(newRemote));
    }

    public void propertySetExistence() {
        assertEquals(cc.getPropertySets().size(), 2, "The config should contain 2 property sets");
        assertTrue(cc.isPropertySetExists("set1"), "The config should contain the property set: 'set1'");
        assertTrue(cc.isPropertySetExists("set2"), "The config should contain the property set: 'set2'");
    }

    public void removePropertySet() {
        PropertySet setToRemove = new PropertySet();
        String setName = "toRemove";
        setToRemove.setName(setName);
        cc.addPropertySet(setToRemove);

        LocalRepoDescriptor localDescriptor = cc.getLocalRepositoriesMap().get("local1");
        localDescriptor.addPropertySet(setToRemove);

        RemoteRepoDescriptor remoteDescriptor = cc.getRemoteRepositoriesMap().get("remote1");
        remoteDescriptor.addPropertySet(setToRemove);

        //Sanity check
        assertNotNull(localDescriptor.getPropertySet(setName), "The descriptor should contain the added " +
                "property set.");

        assertNotNull(cc.removePropertySet(setName), "The property set should have been removed from the central " +
                "config descriptor.");

        assertNull(localDescriptor.getPropertySet(setName), "The property set should have been removed from the " +
                "local repo descriptor.");

        assertNull(remoteDescriptor.getPropertySet(setName),
                "The property set should have been removed from the " +
                        "remote repo descriptor.");
    }

    public void repoLayoutExistence() {
        assertEquals(cc.getRepoLayouts().size(), 2, "The config should contain 2 repository layouts");
        assertTrue(cc.isRepoLayoutExists("layout1"), "The config should contain the repo layout: 'layout1'");
        assertTrue(cc.isRepoLayoutExists("layout2"), "The config should contain the repo layout: 'layout2'");
    }

    public void removeRepoLayout() {
        RepoLayout layoutToRemove = new RepoLayout();
        String name = "toRemove";
        layoutToRemove.setName(name);
        cc.addRepoLayout(layoutToRemove);

        LocalRepoDescriptor localDescriptor = cc.getLocalRepositoriesMap().get("local1");
        localDescriptor.setRepoLayout(layoutToRemove);

        RemoteRepoDescriptor remoteDescriptor = cc.getRemoteRepositoriesMap().get("remote1");
        remoteDescriptor.setRepoLayout(layoutToRemove);

        VirtualRepoDescriptor virtualDescriptor = cc.getVirtualRepositoriesMap().get("virtual1");
        virtualDescriptor.setRepoLayout(layoutToRemove);

        assertNotNull(localDescriptor.getRepoLayout(), "The descriptor should contain the added repo layout");

        assertNotNull(cc.removeRepoLayout(name), "The repo layout should have been removed from the central " +
                "config descriptor.");

        assertEquals(localDescriptor.getRepoLayout(), RepoLayoutUtils.MAVEN_2_DEFAULT,
                "The repo layout should have defaulted.");

        assertEquals(remoteDescriptor.getRepoLayout(), RepoLayoutUtils.MAVEN_2_DEFAULT,
                "The repo layout should have defaulted.");

        assertNull(virtualDescriptor.getRepoLayout(), "The repo layout should have defaulted to null.");
    }

    public void removeKeyPair() {
        // Assert we have the first ome
        assertEquals(cc.getKeyPairs().size(), 1);
        // Add
        KeyPairDescriptor keyPairToRemove = new KeyPairDescriptor();
        String name = "toRemove";
        keyPairToRemove.setKey(name);
        cc.addKeyPair(keyPairToRemove);
        assertEquals(cc.getKeyPairs().size(), 2);

        LocalRepoDescriptor localDescriptor = cc.getLocalRepositoriesMap().get("local1");
        localDescriptor.setPrimaryKeyPairDescriptor(keyPairToRemove);

        VirtualRepoDescriptor virtualDescriptor = cc.getVirtualRepositoriesMap().get("virtual1");
        virtualDescriptor.setPrimaryKeyPairDescriptor(keyPairToRemove);

        assertNotNull(localDescriptor.getPrimaryKeyPairDescriptor(), "The descriptor should contain the added keypair");
        // Validate key that doesnt exist
        cc.removeKeyPair("doesntExist");

        // Act
        cc.removeKeyPair(name);

        assertNull(localDescriptor.getPrimaryKeyPairDescriptor(), "The repo keypair should have defaulted to null.");
        assertNull(virtualDescriptor.getPrimaryKeyPairDescriptor(), "The repo keypair should have defaulted to null.");
        // Assert it was removed
        assertEquals(cc.getKeyPairs().size(), 1);
    }

    public void replicationsExistence() {
        assertNotNull(cc.getRemoteReplication("remote1"));
        assertNotNull(cc.getRemoteReplication("remote2"));
        assertNotNull(cc.getLocalReplication("local1"));
        assertNotNull(cc.getLocalReplication("local2"));
    }

    public void removeReplicationUpdates() {
        RemoteReplicationDescriptor remoteReplication = cc.getRemoteReplication("remote2");
        assertNotNull(remoteReplication, "Expected to find second remote replication.");
        cc.removeRemoteReplication(remoteReplication);
        assertNull(cc.getRemoteReplication("remote2"), "Second remote replication should have been removed.");

        LocalReplicationDescriptor localReplication = cc.getLocalReplication("local2");
        assertNotNull(localReplication, "Expected to find second local replication.");
        cc.removeLocalReplication(localReplication);
        assertNull(cc.getLocalReplication("local2"), "Second local replication should have been removed.");
    }

    @Test
    public void setEnableDisableArchiveIndex() {
        Assert.assertFalse(cc.isArchiveIndexEnabled());
        cc.setArchiveIndexEnabled(true);
        Assert.assertTrue(cc.isArchiveIndexEnabled());
    }

    public void addSystemPropertiesTest() {
        addSystemProperty();
    }

    @Test(expectedExceptions = AlreadyExistsException.class, expectedExceptionsMessageRegExp = "System property" +
            ".*already exists")
    public void addSystemPropertiesTwiceTest() {
        SystemPropertyDescriptor systemPropertyDescriptor = addSystemProperty();
        cc.addSystemProperty(systemPropertyDescriptor);
    }

    public void removeSystemPropertiesTest() {
        SystemPropertyDescriptor systemPropertyDescriptor = addSystemProperty();

        SystemPropertyDescriptor propertyDescriptor = cc.removeSystemProperty("key");
        assertEquals(systemPropertyDescriptor, propertyDescriptor);

        propertyDescriptor = cc.removeSystemProperty("key");
        assertNull(propertyDescriptor);
    }

    public void isSystemPropertiesExistsTest() {
        addSystemProperty();
        assertTrue(cc.isSystemPropertyExists("key"));
        assertTrue(cc.isSystemPropertyExists("key1"));
        assertFalse(cc.isSystemPropertyExists("nonExistingKey"));
    }

    public void setAndGetSystemPropertiesTest() {
        List<SystemPropertyDescriptor> systemPropertyDescriptors = new ArrayList<>();
        systemPropertyDescriptors.add(createSystemProperty("key", "value"));
        systemPropertyDescriptors.add(createSystemProperty("key2", "value2"));
        systemPropertyDescriptors.add(createSystemProperty("key3", "value3"));

        cc.setSystemProperties(systemPropertyDescriptors);
        List<SystemPropertyDescriptor> systemProperties = cc.getSystemProperties();
        assertEquals(systemProperties, systemPropertyDescriptors);
    }

    public void shouldCallEmptyProcessUrlForNonHttpRepoDescriptor() {
        final RemoteRepoDescriptor remoteRepoDescriptor = Mockito.mock(RemoteRepoDescriptor.class);
        remoteRepoDescriptor.setKey("key1");
        cc.addRemoteRepository(remoteRepoDescriptor);
        Mockito.verify(remoteRepoDescriptor).processUrl();
        Mockito.verify(remoteRepoDescriptor, Mockito.times(0)).setUrl(any());
    }

    public void shouldProcessUrlForHttpRepoDescriptor() {
        final RemoteRepoDescriptor remoteRepoDescriptor = Mockito.mock(RemoteRepoDescriptor.class);
        remoteRepoDescriptor.setKey("key2");
        remoteRepoDescriptor.setUrl("testUrl");
        cc.addRemoteRepository(remoteRepoDescriptor);
        Mockito.verify(remoteRepoDescriptor).processUrl();
        Mockito.verify(remoteRepoDescriptor, Mockito.times(1)).setUrl("testUrl");
    }

    public void testGetLocalReplicationsMap() {
        Map<String, Map<String, LocalReplicationDescriptor>> localReplicationsMap = cc.getLocalReplicationsMap();
        assertEquals(localReplicationsMap.size(), cc.getLocalReplications().size());
        List<LocalReplicationDescriptor> localReplications = cc.getLocalReplications();
        localReplications.forEach(localReplicationDescriptor -> {
            Map<String, LocalReplicationDescriptor> perUrlReplication =
                    localReplicationsMap.get(localReplicationDescriptor.getRepoKey());
            assertNotNull(perUrlReplication);
            assertEquals(localReplicationDescriptor, perUrlReplication.get(localReplicationDescriptor.getUrl()));
        });

        LocalReplicationDescriptor localReplicationDescriptor =
                SerializationUtils.clone(cc.getLocalReplications().get(0));
        localReplicationDescriptor.setReplicationKey("cloned");
        cc.addLocalReplication(localReplicationDescriptor);
        Map<String, Map<String, LocalReplicationDescriptor>> localReplicationsMapAfterChange =
                cc.getLocalReplicationsMap();
        assertEquals(localReplicationsMapAfterChange.values().size() + 1, cc.getLocalReplications().size());
        Map<String, LocalReplicationDescriptor> perUrlReplication =
                localReplicationsMapAfterChange.get(localReplicationDescriptor.getRepoKey());
        assertEquals(perUrlReplication.size(), 1);
        assertEquals(perUrlReplication.get(localReplicationDescriptor.getUrl()), localReplicationDescriptor);

        LocalReplicationDescriptor localReplicationDescriptor2 =
                SerializationUtils.clone(cc.getLocalReplications().get(0));
        localReplicationDescriptor2.setReplicationKey("cloned2");
        localReplicationDescriptor2.setUrl("http://copied.com");
        cc.addLocalReplication(localReplicationDescriptor2);
        localReplicationsMapAfterChange = cc.getLocalReplicationsMap();
        perUrlReplication = localReplicationsMapAfterChange.get(localReplicationDescriptor.getRepoKey());
        assertEquals(perUrlReplication.size(), 2);
        assertEquals(perUrlReplication.get(localReplicationDescriptor2.getUrl()), localReplicationDescriptor2);

        cc.removeLocalReplication(localReplicationDescriptor2);
        cc.removeLocalReplication(localReplicationDescriptor);
    }

    @NotNull
    private SystemPropertyDescriptor addSystemProperty() {
        SystemPropertyDescriptor systemPropertyDescriptor = createSystemProperty("key", "Value");
        cc.addSystemProperty(systemPropertyDescriptor);
        assertTrue(cc.getSystemProperties().contains(systemPropertyDescriptor));
        return systemPropertyDescriptor;
    }

    @NotNull
    private SystemPropertyDescriptor createSystemProperty(String key, String value) {
        return new SystemPropertyDescriptor(key, value);
    }

    private void addDefaultProxyToRemoteRepositories(ProxyDescriptor proxyDescriptor) {
        Map<String, RemoteRepoDescriptor> descriptorOrderedMap = cc.getRemoteRepositoriesMap();
        for (RemoteRepoDescriptor descriptor : descriptorOrderedMap.values()) {
            if (descriptor instanceof HttpRepoDescriptor) {
                HttpRepoDescriptor httpRepoDescriptor = (HttpRepoDescriptor) descriptor;
                httpRepoDescriptor.setProxy(proxyDescriptor);
            }
        }
    }

    private void addDefaultProxyToFederatedRepositories(ProxyDescriptor proxyDescriptor) {
        Map<String, FederatedRepoDescriptor> descriptorOrderedMap = cc.getFederatedRepositoriesMap();
        for (FederatedRepoDescriptor descriptor : descriptorOrderedMap.values()) {
            descriptor.setProxy(proxyDescriptor);
        }
    }

    @DataProvider
    private static Object[][] proxyKeyProvider() {
        return new Object[][] {
                {"", false},
                {"  ", false},
                {"1foo", false},
                {"1", false},
                {"123", false},
                {"$key", false},
                {"key&", false},
                {"key?", false},
                {"key*", false},
                {"key/", false},
                {"key\\", false},
                {"foo", true},
                {"foo-bar", true},
                {"_bar", true}
        };
    }
}

