package org.artifactory.core.api.ph.service;

import org.artifactory.core.api.ph.model.PackageArtifact;

import java.util.Map;
import java.util.Optional;
import java.util.stream.Stream;

public interface PackageSearchArtifactService {

    /**
     * Find artifacts by extension(s).
     * Path is optional. If provided, it will search by exact path and underneath it.
     * Only artifacts with read permissions are returned.
     * Use case example: {@code ("deb-local", "", "deb", "ddeb")})
     *
     * @param extensions The extensions of the artifacts (without '.', e.g. {@code "zip", "tar.gz"})
     * @param path       The path of the artifact (without leading '/', e.g. root is {@code ""})
     * @return a {@link Stream} with the {@link PackageArtifact}s
     */
    Stream<PackageArtifact> findArtifactsByExtension(String repoKey, String path, String... extensions);

    /**
     * Search by exact path and underneath it.
     * Only artifacts with read permissions are returned.
     * Use case example: {@code ("deb-local", "pool")})
     *
     * @param path The path of the artifact (without leading '/', e.g. root is {@code ""})
     * @return a {@link Stream} with the {@link PackageArtifact}s
     */
    Stream<PackageArtifact> findArtifactsChildren(String repoKey, String path);

    /**
     * Search by exact path match and underneath it + By artifact name prefix.
     * Only artifacts with read permissions are returned.
     * Use case example: {@code ("cran-local", "sources/src/contrib", "my-package_")})
     *
     * @param path The path of the artifact (without leading '/', e.g. root is {@code ""})
     * @return a {@link Stream} with the {@link PackageArtifact}s
     */
    Stream<PackageArtifact> findArtifactsChildrenStartWithName(String repoKey, String path, String name);

    /**
     * Search for artifacts in any depth that their names are equal to the given name.
     * Only artifacts with read permissions are returned.
     *
     * @param name The name of the artifact (without astrix, e.g. {@code "super.zip"})
     * @return a {@link Stream} with the {@link PackageArtifact}s
     */
    Stream<PackageArtifact> findArtifactsByExactName(String repoKey, String name);

    /**
     * Search for artifacts by exact match to path and name.
     * Only artifacts with read permissions are returned.
     *
     * @param path The path of the artifact (without astrix or leading '/', e.g. root is {@code ""})
     * @param name The name of the artifact (without astrix, e.g. {@code "super.zip"})
     * @return a {@link PackageArtifact} that match to the searched path and name
     */
    Optional<PackageArtifact> findArtifactByExactPathAndName(String repoKey, String path, String name);

    /**
     * Search for artifacts with the existence of the given prop keys.
     * Optional filter by exact path match and underneath it.
     * Only artifacts with read permissions are returned.
     * Use case example: {@code ("deb-local", "deb.architecture", "deb.distribution", "deb.component")}
     *
     * @param path     The path of the artifact (without astrix or leading '/', e.g. root is {@code ""})
     * @param propKeys A set of prop keys to search for (e.g. {@code "debian.architecture"})
     * @return a {@link Stream} with the {@link PackageArtifact}s
     */
    Stream<PackageArtifact> findArtifactsByPropKeys(String repoKey, String path, String... propKeys);

    /**
     * Search for artifacts by exact match to the given prop keys and values.
     * Optional filter by exact path match and underneath it.
     * Only artifacts with read permissions are returned.
     * Use case example: {@code ("deb-local", {"deb.architecture", "amd64"})}
     *
     * @param path       The path of the artifact (without astrix or leading '/', e.g. root is {@code ""})
     * @param properties A map of properties to search by (e.g. {@code "{debian.architecture , amd64}"})
     * @return a {@link Stream} with the {@link PackageArtifact}s
     */
    Stream<PackageArtifact> findArtifactsByPropKeyAndVal(String repoKey, String path, Map<String, String> properties);

}
