package org.artifactory.storage.db.remotebintask;

import org.apache.commons.lang3.StringUtils;
import org.artifactory.common.config.db.ArtifactoryDbProperties;
import org.artifactory.storage.db.remotebintask.entity.RemoteBinaryTask;
import org.artifactory.storage.db.util.BaseDao;
import org.artifactory.storage.db.util.JdbcHelper;
import org.jfrog.storage.DbType;
import org.jfrog.storage.util.querybuilder.QueryWriter;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Repository;
import org.springframework.util.Assert;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

import static java.util.concurrent.TimeUnit.SECONDS;
import static org.artifactory.common.ConstantValues.haHeartbeatStaleIntervalSecs;

@Repository
public class RemoteBinaryTaskDao extends BaseDao {

    private static final Logger log = LoggerFactory.getLogger(RemoteBinaryTaskDao.class);

    public static final String COLUMNS =
            "id, sha1, time, local_repo_key, remote_url_hash, start_time, count, server_id";

    public static final String TABLE_NAME = "binaries_tasks";

    private static final String INSERT_SQL = "INSERT INTO " + TABLE_NAME + " (" + COLUMNS + ") " +
            "VALUES (?, ?, ?, ?, ?, ?, ?, ?)";

    private static final String SELECT_ALL_SQL = "SELECT " + COLUMNS + " FROM " + TABLE_NAME;

    private static final String SELECT_BY_ID_SQL = SELECT_ALL_SQL + " WHERE id = ?";
    private static final String DELETE_BY_ID_SQL = "DELETE FROM " + TABLE_NAME + " WHERE id = ?";
    private static final String ALLOCATE_SQL =
            "UPDATE " + TABLE_NAME + " SET start_time=?, count=count+1, server_id=? WHERE id=?";
    private static final String DEALLOCATE_BY_ID_SQL =
            "UPDATE " + TABLE_NAME + " SET start_time=NULL, server_id=NULL WHERE id=?";
    private static final String DEALLOCATE_BY_SERVER_ID_SQL =
            "UPDATE " + TABLE_NAME + " SET start_time=NULL, server_id=NULL WHERE server_id=?";

    private final DbType dbType;

    public RemoteBinaryTaskDao(JdbcHelper jdbcHelper, ArtifactoryDbProperties dbProperties) {
        super(jdbcHelper);
        this.dbType = dbProperties.getDbType();
    }

    public void add(long id, @Nonnull String sha1, @Nonnull String localRepoKey, @Nullable String remoteUrl)
            throws SQLException {
        Assert.notNull(sha1, "sha1");
        Assert.notNull(localRepoKey, "localRepoKey");
        jdbcHelper.executeUpdate(INSERT_SQL, id, sha1, System.currentTimeMillis(),
                localRepoKey, StringUtils.isBlank(remoteUrl) ? null : remoteUrl.hashCode(),
                null, 0, null);
        log.debug("Created RemoteBinaryTask sha1: {}, localRepoKey: {}, remoteUrl: {}", sha1, localRepoKey,
                remoteUrl);
    }

    public void allocate(long id, @Nonnull String serverId) throws SQLException {
        Assert.notNull(serverId, "serverId");

        log.debug("Allocating RemoteBinaryTask id: {}, serverId: {}", id, serverId);
        jdbcHelper.executeUpdate(ALLOCATE_SQL, System.currentTimeMillis(), serverId, id);
    }

    public void deAllocateById(long id) throws SQLException {
        log.debug("DeAllocating RemoteBinaryTask id: {}", id);
        jdbcHelper.executeUpdate(DEALLOCATE_BY_ID_SQL, id);
    }

    //todo: call on server startup
    public void deAllocateByServerId(@Nonnull String serverId) throws SQLException {
        Assert.notNull(serverId, "serverId");
        log.debug("Resetting RemoteBinaryTasks serverId: {}", serverId);
        jdbcHelper.executeUpdate(DEALLOCATE_BY_SERVER_ID_SQL, serverId);
    }

    public void remove(long id) throws SQLException {
        log.debug("Deleting RemoteBinaryTask id: {}", id);
        jdbcHelper.executeUpdate(DELETE_BY_ID_SQL, id);
    }

    public @Nullable RemoteBinaryTask getById(long id) throws SQLException {
        try (ResultSet rs = jdbcHelper.executeSelect(SELECT_BY_ID_SQL, id)) {
            if (rs.next()) {
                return task(rs);
            } else {
                return null;
            }
        }
    }

    public @Nonnull List<RemoteBinaryTask> getAll() throws SQLException {
        List<RemoteBinaryTask> tasks = new ArrayList<>();
        try (ResultSet rs = jdbcHelper.executeSelect(SELECT_ALL_SQL)) {
            while (rs.next()) {
                tasks.add(task(rs));
            }
        }
        return tasks;
    }

    public @Nullable List<RemoteBinaryTask> getTasks(@Nonnull String serverId, long batchSize,
            int maxExecutionsCount, long retryDelayMs,
            long maximumExecTimeMs) throws SQLException {
        String nextSql = getNextBatchQuery(batchSize);
        List<RemoteBinaryTask> tasks = new ArrayList<>();
        Assert.notNull(serverId, "serverId");
        RemoteBinaryTask task;
        long now = System.currentTimeMillis();
        try (ResultSet rs = jdbcHelper.executeSelect(nextSql, maxExecutionsCount, retryDelayMs, now, maximumExecTimeMs,
                now, SECONDS.toMillis(haHeartbeatStaleIntervalSecs.getInt()), now)) {
            while (rs.next()) {
                task = task(rs);
                allocate(task.getId(), serverId);
                tasks.add(task);
            }
        }
        return tasks;
    }

    //also checks artifactory servers table status
    private String getNextBatchQuery(long batchSize) {
        String nextSql = new QueryWriter(this.dbType).select(COLUMNS).from(TABLE_NAME)
                .where("count < ? " +
                        "AND (" +
                        "(server_id IS NULL AND time + (count * ?) <= ?) " +
                        "OR" +
                        "(server_id IS NOT NULL " +
                        "AND (" +
                        "(start_time + ? < ?) " +
                        "OR" +
                        "(server_id NOT IN " +
                        "(SELECT server_id FROM artifactory_servers " +
                        "WHERE last_heartbeat + ? >= ?)" +
                        ")" +
                        ")" +
                        ")" +
                        ")")
                .orderBy("time")
                .limit(batchSize)
                .build();
        if (DbType.POSTGRESQL.equals(dbType)) {
            nextSql = nextSql + " FOR UPDATE SKIP LOCKED";
        } else if (DbType.MYSQL.equals(dbType) || DbType.MARIADB.equals(dbType)) {
            nextSql = nextSql + " FOR UPDATE";
        }
        return nextSql;
    }

    private RemoteBinaryTask task(ResultSet rs) throws SQLException {
        return RemoteBinaryTask.builder()
                .id(rs.getLong("id"))
                .sha1(rs.getString("sha1"))
                .time(rs.getLong("time"))
                .localRepoKey(rs.getString("local_repo_key"))
                .remoteUrlHash(rs.getInt("remote_url_hash"))
                .startTime(rs.getObject("start_time", Long.class))
                .count(rs.getInt("count"))
                .serverId(rs.getString("server_id"))
                .build();
    }


}
