/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.ui.rest.service.artifacts.browse.generic;

import com.google.common.collect.Lists;
import org.apache.http.HttpStatus;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.CoreAddons;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.api.repo.BaseBrowsableItem;
import org.artifactory.api.repo.BrowsableItemCriteria;
import org.artifactory.api.repo.RepositoryBrowsingService;
import org.artifactory.api.repo.RootNodesFilterResult;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.common.ConstantValues;
import org.artifactory.md.Properties;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.RepoPathFactory;
import org.artifactory.repo.service.InternalRepositoryService;
import org.artifactory.repo.virtual.VirtualRepo;
import org.artifactory.rest.common.service.RestResponse;
import org.artifactory.rest.common.service.RestService;
import org.artifactory.ui.utils.BrowseUtils;
import org.joda.time.format.DateTimeFormat;
import org.joda.time.format.DateTimeFormatter;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.MediaType;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.Serializable;
import java.util.Collection;
import java.util.Comparator;
import java.util.List;
import java.util.Locale;
import java.util.stream.Stream;

/**
 * @author Chen Keinan
 */
public abstract class BaseBrowseNativeService implements RestService<String> {

    protected static final DateTimeFormatter DATE_FORMAT = DateTimeFormat.forPattern("dd-MMM-yyyy HH:mm")
            .withLocale(Locale.ENGLISH);
    protected List<? extends BaseBrowsableItem> items;
    protected int columnSize;
    protected boolean printParent;
    public static final String REALM = "Artifactory Realm";
    public static final String ATTR_ARTIFACTORY_REPOSITORY_PATH = "artifactory.repository_path";
    public static final String ATTR_ARTIFACTORY_REQUEST_PROPERTIES = "artifactory.request_properties";
    public String head;
    @Autowired
    protected InternalRepositoryService repositoryService;
    @Autowired
    protected RepositoryBrowsingService repoBrowsingService;
    @Autowired
    protected AuthorizationService authorizationService;
    @Autowired
    CentralConfigService centralConfigService;

    /**
     * sort retrieved browsable item from repository
     *
     * @param repoPath              - repository path
     * @param requestProps          - request properties
     * @param rootNodesFilterResult - object hold the List node acceptance status flag,if flagged to false meaning
     *                              at least one node has read permission issue by user
     * @return list of sorted browsable item
     */
    protected List<? extends BaseBrowsableItem> getItems(RepoPath repoPath, Properties requestProps,
            RootNodesFilterResult rootNodesFilterResult) {
        boolean updateRootNodesFilterFlag = authorizationService.isAnonymous();
        List<? extends BaseBrowsableItem> items =
                getBrowsableItemsList(repoPath, requestProps, rootNodesFilterResult, updateRootNodesFilterFlag);
        items.sort(new ItemInfoComparator());
        return items;
    }

    /**
     * get Browsable Item from repositories (local/remote/virtual) based on search criteria
     *
     * @param repoPath                  - repository path
     * @param requestProps              - properties of the request
     * @param rootNodesFilterResult     - object hold the List node acceptance status flag,if flagged to false meaning
     *                                  at least one node has read permission issue by user
     * @param updateRootNodesFilterFlag - if true keep the acceptance status flag
     */
    private List<? extends BaseBrowsableItem> getBrowsableItemsList(RepoPath repoPath, Properties requestProps,
            RootNodesFilterResult rootNodesFilterResult, boolean updateRootNodesFilterFlag) {
        List<BaseBrowsableItem> items = Lists.newArrayList();
        BrowsableItemCriteria.Builder builder = new BrowsableItemCriteria.Builder(repoPath).
                requestProperties(requestProps);
        boolean includeChecksums = !ConstantValues.uiHideChecksums.getBoolean();
        builder.includeChecksums(includeChecksums);
        BrowsableItemCriteria criteria = builder.build();
        if (repositoryService.remoteRepoDescriptorByKey(repoPath.getRepoKey()) != null) {
            items = repoBrowsingService
                    .getRemoteRepoBrowsableChildren(criteria, updateRootNodesFilterFlag, rootNodesFilterResult);
        } else if (repositoryService.localFederatedCachedRepoDescriptorByKey(repoPath.getRepoKey()) != null) {
            items = repoBrowsingService
                    .getLocalRepoBrowsableChildren(criteria, updateRootNodesFilterFlag, rootNodesFilterResult);
        } else if (repositoryService.virtualRepoDescriptorByKey(repoPath.getRepoKey()) != null) {
            items = repoBrowsingService
                    .getVirtualRepoBrowsableChildren(criteria, updateRootNodesFilterFlag, rootNodesFilterResult);
        }
        if (items != null && !includeChecksums) {
            items = BrowseUtils.filterChecksums(items);
        }
        return items;
    }

    /**
     * set response to http challenge message to anonymous if node list is empty due to missing permission
     *
     * @param repoPath              - repo path
     * @param rootNodesFilterResult - object hold empty list canRead flag
     * @param items                 - list of Items
     * @return true if challenge message was set, false otherwise
     */
    protected boolean setResponseToHttpChallenge(RepoPath repoPath, RootNodesFilterResult rootNodesFilterResult,
            List<? extends BaseBrowsableItem> items, RestResponse response, boolean isNativeModelRequest) {

        if (isItemListEmptyDueAnonymousMissingPermission(repoPath, rootNodesFilterResult, items)) {
            HttpServletResponse servletResponse = response.getServletResponse();
            if (!isNativeModelRequest) {
                servletResponse.addHeader("WWW-Authenticate", "Basic realm=\"" + REALM + "\"");
                servletResponse.setContentType(MediaType.APPLICATION_JSON_VALUE);
            }
            int statusCode = HttpStatus.SC_UNAUTHORIZED;
            response.responseCode(statusCode);
            response.error("Browse request for repo:path " + repoPath + " is forbidden " +
                    "for user 'anonymous'.");
            return true;
        }
        return false;
    }

    /**
     * @param rootNodesFilterResult - object hold List canRead flag
     * @param items                 list of items
     * @return true if list is empty due to anonymous missing permission to read, false otherwise
     */
    protected boolean isItemListEmptyDueAnonymousMissingPermission(RepoPath repoPath,
            RootNodesFilterResult rootNodesFilterResult,
            List<? extends BaseBrowsableItem> items) {
        return items.isEmpty() && (!rootNodesFilterResult.isAllItemNodesCanRead() || !hasPermissionOnPath(repoPath));
    }

    protected boolean hasPermissionOnPath(RepoPath repoPath) {
        if (repositoryService.virtualRepoDescriptorByKey(repoPath.getRepoKey()) != null) {
            VirtualRepo virtualRepo = repositoryService.virtualRepositoryByKey(repoPath.getRepoKey());
            if (virtualRepo != null) {
                return Stream.of(virtualRepo.getResolvedLocalAndFederatedRepos(), virtualRepo.getResolvedRemoteRepos())
                        .flatMap(Collection::stream)
                        .anyMatch(repo -> authorizationService
                                .canRead(RepoPathFactory.create(repo.getKey(), repoPath.getPath())));
            }
            return false;
        } else {
            return authorizationService.canRead(repoPath);
        }
    }

    /**
     * get addresses footer
     *
     * @param request - http servlet request
     * @return - address footer
     */
    protected String getAddressFooter(HttpServletRequest request) {
        StringBuilder addressBuilder = new StringBuilder();
        String versionInfo = getListBrowsingVersion();
        String serverName = request.getServerName();
        addressBuilder
                .append(String.format("%s Server at %s Port %s", versionInfo, serverName, request.getServerPort()));
        return addressBuilder.toString();
    }

    /**
     * get List browsing versions
     *
     * @return - list browsing versions
     */
    public String getListBrowsingVersion() {
        CoreAddons addon = ContextHelper.get().beanForType(AddonsManager.class).addonByType(CoreAddons.class);
        return addon.getListBrowsingVersion();
    }

    /**
     * item comparator for sorting
     */
    private static class ItemInfoComparator implements Comparator<BaseBrowsableItem>, Serializable {
        @Override
        public int compare(BaseBrowsableItem o1, BaseBrowsableItem o2) {
            final int folderCmp = Boolean.compare(o2.isFolder(), o1.isFolder());
            if (folderCmp != 0) {
                return folderCmp;
            }
            return o1.getName().compareToIgnoreCase(o2.getName());
        }
    }
}

