package org.artifactory.ui.rest.service.admin.configuration.general;

import lombok.NonNull;
import org.apache.commons.imaging.ImageReadException;
import org.apache.commons.io.FileUtils;
import org.apache.http.HttpStatus;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.descriptor.config.MutableCentralConfigDescriptor;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.ArtifactoryRestResponse;
import org.artifactory.ui.rest.model.utils.FileUpload;
import org.glassfish.jersey.media.multipart.FormDataBodyPart;
import org.glassfish.jersey.media.multipart.FormDataMultiPart;
import org.jfrog.common.ResourceUtils;
import org.mockito.Mock;
import org.mockito.MockedStatic;
import org.mockito.Mockito;
import org.mockito.MockitoAnnotations;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.Test;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.List;
import java.util.Map;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertTrue;

public class UploadLogoServiceTest {

    private final UploadLogoService service = new UploadLogoService();
    @Mock
    ArtifactoryContext artifactoryContext;
    @Mock
    ArtifactoryHome artifactoryHome;
    @Mock
    CentralConfigService centralConfigService;

    private AutoCloseable autoCloseable;

    @BeforeClass
    public void setUp() {
        autoCloseable = MockitoAnnotations.openMocks(this);
        ArtifactoryContextThreadBinder.bind(artifactoryContext);
        when(artifactoryContext.getArtifactoryHome()).thenReturn(artifactoryHome);
        when(artifactoryHome.getTempWorkDir()).thenReturn(new File("/tmp/data/"));
        service.centralConfigService = centralConfigService;
    }

    public void tearDown() throws Exception {
        ArtifactoryContextThreadBinder.unbind();
        autoCloseable.close();
    }

    @Test
    public void isARealImage() throws Exception {
        File rceFile = ResourceUtils.getResourceAsFile("/uiresource.malicious/jfrog.png");
        byte[] fileBytes = FileUtils.readFileToByteArray(rceFile);
        ArtifactoryRestResponse<Object> response = new ArtifactoryRestResponse<>() {
        };
        ArtifactoryRestRequest<Object> request = Mockito.mock(ArtifactoryRestRequest.class);
        setMockForReadFile(fileBytes, request);
        MockedStatic<Files> filesMockedStatic = Mockito.mockStatic(Files.class);
        filesMockedStatic.when(() -> Files.move(any(), any(), any())).thenReturn(null);
        filesMockedStatic.when(() -> Files.readAllBytes(any())).thenReturn(fileBytes);
        filesMockedStatic.when(() -> Files.createTempFile("imageio", ".tmp")).thenReturn(Paths.get("tmp/"));

        service.execute(request, response);
        assertTrue((response.buildResponse().getStatus() == HttpStatus.SC_OK));
        filesMockedStatic.close();
    }

    @Test
    public void isNonImageFile() throws Exception {
        File rceFile = ResourceUtils.getResourceAsFile("/browse/treebrowser/test.zip");
        ArtifactoryRestResponse<Object> response = new ArtifactoryRestResponse<>() {
        };
        byte[] fileBytes = FileUtils.readFileToByteArray(rceFile);
        ArtifactoryRestRequest<Object> request = Mockito.mock(ArtifactoryRestRequest.class);
        setMockForReadFile(fileBytes, request);

        MockedStatic<Files> filesMockedStatic = Mockito.mockStatic(Files.class);
        filesMockedStatic.when(() -> Files.move(any(), any(), any())).thenReturn(null);
        filesMockedStatic.when(() -> Files.readAllBytes(any())).thenReturn(fileBytes);
        filesMockedStatic.when(() -> Files.createTempFile("imageio", ".tmp")).thenReturn(Paths.get("tmp/"));

        service.execute(request, response);
        assertTrue(((String) response.buildResponse().getEntity()).contains("Uploaded logo is not an image file"));
        assertEquals(HttpStatus.SC_BAD_REQUEST, response.buildResponse().getStatus());
        filesMockedStatic.close();
    }

    @Test
    public void imageContainsEmbeddedRCEInJpgFile() throws Exception {
        File rceFile = ResourceUtils.getResourceAsFile("/uiresource.malicious/POC_phpinfo-metadata.jpeg");
        ArtifactoryRestResponse<Object> response = new ArtifactoryRestResponse<>() {
        };
        byte[] fileBytes = FileUtils.readFileToByteArray(rceFile);
        ArtifactoryRestRequest<Object> request = Mockito.mock(ArtifactoryRestRequest.class);
        setMockForReadFile(fileBytes, request);

        MockedStatic<Files> filesMockedStatic = Mockito.mockStatic(Files.class);
        filesMockedStatic.when(() -> Files.move(any(), any(), any())).thenReturn(null);
        filesMockedStatic.when(() -> Files.readAllBytes(any())).thenReturn(fileBytes);
        filesMockedStatic.when(() -> Files.createTempFile("imageio", ".tmp")).thenReturn(Paths.get("tmp/"));

        service.execute(request, response);
        assertTrue(
                ((String) response.buildResponse().getEntity())
                        .contains("Image contains potentially malicious code, check logs for details"));
        assertEquals(HttpStatus.SC_BAD_REQUEST, response.buildResponse().getStatus());
        filesMockedStatic.close();
    }

    @Test(expectedExceptions = ImageReadException.class)
    public void testIsExifMalicious() throws IOException, ImageReadException {
        service.isExifMalicious(service.getImageBytes(
                ResourceUtils.getResourceAsFile("/uiresource.malicious/POC_phpinfo-metadata.jpeg")), new File("bla"));
    }

    private void setMockForReadFile(@NonNull byte[] fileBytes,
            @NonNull ArtifactoryRestRequest request)
            throws IOException {
        FileUpload fileUpload = Mockito.mock(FileUpload.class);
        FormDataMultiPart formDataMultiPart = Mockito.mock(FormDataMultiPart.class);
        FormDataBodyPart formDataBodyPart = Mockito.mock(FormDataBodyPart.class);
        MutableCentralConfigDescriptor descriptor = Mockito.mock(MutableCentralConfigDescriptor.class);
        Map<String, List<FormDataBodyPart>> fields = Map.of("file", List.of(formDataBodyPart));

        Mockito.when(request.getImodel()).thenReturn(fileUpload);
        Mockito.when(fileUpload.getFormDataMultiPart()).thenReturn(formDataMultiPart);
        Mockito.when(formDataMultiPart.getFields()).thenReturn(fields);
        Mockito.when(formDataBodyPart.getValueAs(byte[].class)).thenReturn(fileBytes);
        Mockito.when(centralConfigService.getMutableDescriptor()).thenReturn(descriptor);
        Mockito.when(descriptor.getFileUploadMaxSizeMb()).thenReturn(100);
        Mockito.when(artifactoryHome.getLogoDir()).thenReturn(new File("/tmp/logo"));
    }
}